Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomBridge.
Require Import DomSet.
Require Import DomMap.
Require Import DomZ.
Require Import DomItv.
Require Import DomAbsValue.
Require Import DomAcc.

Require Import Syntax.

Require Import Containers.SetProperties.
Require Import Containers.MapInterface.

Set Implicit Arguments.


Section Memories.
  Definition memT := withtop Map![locT, valueT].
  Global Instance memT_setoid : setoid memT := _.
  Global Instance memT_setoidDec : @setoidDec memT _ := _.
  Global Instance memT_poset : @poset memT _ := _.
  Global Instance memT_posetDec : @posetDec memT _ _ _ := _. 
  Global Instance memT_hasTop : @hasTop memT _ _ := _.
  Global Instance memT_hasJoin : @hasJoin memT _ _ := _.
  Global Instance memT_maplike : maplike (Key:=locT) (Elt:=valueT) (M:=memT) := _.
  Global Instance memT_restricting_maplike : restricting_maplike (K:=locT) (E:=valueT) (M:=memT) := _.
  Global Instance memT_restricting_join_maplike : restricting_join_maplike (K:=locT) (E:=valueT) (M:=memT) := _.

  Lemma mem_join_spec (l:locT) (m n:memT) :
    (mlookup (Key:=locT) (Elt:=valueT) (M:=memT) l (m ⊔ n))
    ≡ (mlookup (Key:=locT) (Elt:=valueT) (M:=memT) l m)
    ⊔ (mlookup (Key:=locT) (Elt:=valueT) (M:=memT) l n).
  Proof.
    move: m n=> [m|] [n|] /=; dom_tac.
    reflexivity.
    by symmetry; apply: join_top_rhs.
    by symmetry; apply: join_top_lhs.
    by symmetry; apply: join_top_lhs.
  Qed.

  Lemma mem_eq_spec (m n:memT) (Hm : m <> top) (Hn : n <> top)
        (Heq : forall (loc:locT), eq m[loc]! n[loc]!) :
    eq m n.
  Proof. by move: m n Hm Hn Heq=> [m|] [n|] Hm Hn Heq. Qed.

  Lemma mem_le_spec (m n:memT) (Hm : m <> top) (Hn : n <> top)
        (Hle : forall (loc:locT), le m[loc]! n[loc]!) :
    le m n.
  Proof. by move: m n Hm Hn Hle=> [m|] [n|] Hm Hn Hle. Qed.

  Lemma mem_mrestrict_def (ls: \powerset locT) (m:Map![locT, valueT]) :
    mrestrict ls (Elt m) = Elt (mrestrict (E:=valueT) ls m).
  Proof. reflexivity. Qed.

  Lemma mem_mrestrict_join_def (ls: \powerset locT) (m n:Map![locT, valueT]) :
    mrestrict_join ls (Elt m) (Elt n) = Elt (mrestrict_join (E:=valueT) ls m n).
  Proof. reflexivity. Qed.

  Lemma mem_mrestrict_top (locs: \powerset locT) :
    eq (mrestrict locs (top:memT)) top.
  Proof.
    rewrite/mrestrict/=.
    reflexivity.
  Qed.

  Lemma mem_mrestrict_join_top_rhs (locs: \powerset locT) (mem:memT) :
    eq (mrestrict_join locs mem top) top.
  Proof.
    rewrite/mrestrict_join.
    rewrite mem_mrestrict_top.
    exact: join_top_rhs.
  Qed.

  Lemma mem_mrestrict_join_top_lhs (locs: \powerset locT) (mem:memT) :
    eq (mrestrict_join locs top mem) top.
  Proof.
    rewrite/mrestrict_join.
    rewrite mem_mrestrict_top.
    exact: join_top_lhs.
  Qed.
  
  Lemma mem_mrestrict_spec (ls: \powerset locT) (m:memT) (loc:locT) :
    eq (mrestrict ls m)[loc]!
       (if m is Elt m'
        then if loc !\in ls
             then m[loc]!
             else bot
        else top).
  Proof.
    move: m=>[m|]; last done.
    exact: map_mrestrict_spec.
  Qed.

  Lemma mem_mrestrict_join_spec (ls: \powerset locT) (m n:memT) (loc:locT) :
    eq (mrestrict_join ls m n)[loc]!
       (if m is Elt m'
        then if n is Elt n'
             then if loc !\in ls
                  then m[loc]!
                  else n[loc]!
             else top
        else top).
  Proof.
    move: m n=>[m|] [n|] //=; dom_tac.
    case: (_ !\in _); rewrite/negb.
    by rewrite join_bot; reflexivity.
    by rewrite join_commutative join_bot; reflexivity.
  Qed.

  Lemma mrestrict_join_mlookup : forall (l:locT) (ls: \powerset locT) (m m':memT)
    (Hl:l !\in ls) (Hnottop: not (m' ≡ top)),
    (mlookup (Key:=locT) (Elt:=valueT) (M:=memT) l
      (mrestrict_join (K:=locT) (E:=valueT) (M:=memT) ls m m')) 
    ≡ mlookup (Key:=locT) (Elt:=valueT) (M:=memT) l m.
  Proof.
    intros. rewrite mem_join_spec.
    rewrite <- mrestrict_lookup_spec; auto.
    destruct m' as [m'|].
    unfold mrestrict,memT_restricting_maplike,withtop_restricting_maplike.
    unfold mlookup at 2,memT_maplike,withtop_maplike.
    dom_tac.
    rewrite Hl /negb join_bot.
    reflexivity.
    by contradict Hnottop; reflexivity.
  Qed.
  
  Variable (prog:progT) (acc:accT).

  Definition memory_lookup (ls:locsT) (mem:memT) : valueT :=
    if ls is Elt ls'
    then SetInterface.fold
           (fun loc val => join mem[loc]! val)
           ls'
           bot
    else top.

  Definition memory_update (ls:locsT) (value:valueT) (mem:memT) : memT :=
    if mem is Elt mem'
    then if ls is Elt ls'
         then Elt
                (let ls'' := SetInterface.elements ls' in
                 if ls'' is loc::nil
                 then if is_strong acc loc
                      then mem'[loc <- value]!
                      else mem'[loc <- join mem'[loc]! value]!
                 else fold_left
                        (fun mem' loc => mem'[loc <- join mem'[loc]! value]!)
                        ls''
                        mem')
         else top
    else top.
  
  Lemma memory_update_spec (ls:locsT) (value:valueT) (mem:memT) (loc:locT) :
    eq (memory_update ls value mem)[loc]!
       (if mem is Elt mem'
        then if ls is Elt ls'
             then if loc !\in ls'
                  then if eqb (SetInterface.cardinal ls') 1 && is_strong acc loc
                       then value
                       else join mem[loc]! value
                  else mem[loc]!
             else top
        else top).
  Proof.
    rewrite/memory_update.
    move: mem=> [mem|]; last reflexivity.
    move: ls=> [ls|]; last reflexivity.
    rewrite/smem/set_setlike.
    rewrite SetFacts.elements_b SetInterface.cardinal_1.
    move: (SetInterface.elements ls)=> {ls} ls.
    rewrite [(Elt _)[_]!]/mlookup/memT_maplike/withtop_maplike.
    move: ls=> [|loc' ls]; first by reflexivity.
    move: ls=> [|loc'' ls].
    { rewrite/length eqb_refl /andb.
      case Heqb: (eqb loc loc').
      { move/eqbP in Heqb; usual_tac Heqb.
        rewrite/existsb is_compare_eqb eqb_refl /orb.
        by case: (is_strong _ _); rewrite map_madd_spec eqb_refl; reflexivity.
      }
      { rewrite/existsb is_compare_eqb Heqb /orb.
        by case: (is_strong _ _); rewrite map_madd_spec eqb_sym Heqb; reflexivity.
      }
    }
    { rewrite [eqb _ 1]/eqb.
      rewrite [eqb_ _ 1]/=/andb.
      move: (_ :: _ :: _)=> {loc' loc'' ls} ls.
      rewrite <- fold_left_rev_right.
      rewrite existsb_rev.
      move: (rev ls)=> {ls} ls.
      elim: ls mem=> [|loc' ls IH] mem; first reflexivity.
      case Heqb: (eqb loc loc'); rewrite/existsb is_compare_eqb Heqb /orb.
      { move/eqbP in Heqb; usual_tac Heqb.
        rewrite/fold_right IH.
        rewrite map_madd_spec eqb_refl.
        case: (existsb _ _); last reflexivity.
        apply/DomType.le_antisym.
        { apply: join_lub.
          exact/DomType.eq_le/eq_refl.
          exact: join_rhs.
        }
        { exact: join_lhs. }
      }
      { rewrite/fold_right.
        rewrite map_madd_spec eqb_sym Heqb IH.
        reflexivity.
      }
    }
  Qed.

  Lemma memory_lookup_fold_spec :
    forall (locs:list locT) (lhs rhs:valueT) (mem:memT),
      eq (fold_left (λ (a : valueT) (e : locT), (mem) [e ]! ⊔ a)
                    locs (lhs ⊔ rhs))
         (lhs ⊔ (fold_left (λ (a : valueT) (e : locT), (mem) [e ]! ⊔ a)
                           locs rhs)).
  Proof.
    move=> locs lhs rhs mem.
    rewrite <- 2! fold_left_rev_right.
    elim: (rev locs) lhs rhs => [|lhd ltl IH] lhs rhs.
    { simpl; exact:eq_refl. }
    { rewrite/fold_right.
      rewrite <- join_associative.
      rewrite (join_commutative lhs (mem[lhd]!)).
      rewrite join_associative.
      apply join_mor; first by exact:eq_refl.
      rewrite/fold_right in IH.
      apply (IH lhs rhs).
    }
  Qed.

  Lemma memory_lookup_ub :
    forall (locs:Set![locT]) (mem:memT) (loc:locT)
           (Hloc: loc !\in locs),
      le mem[loc]! (memory_lookup (Elt locs) mem).
  Proof.
    move=> locs mem loc Hlocin.
    rewrite/memory_lookup.
    rewrite -> SetInterface.fold_1.
    simpl in Hlocin.
    apply SetFacts.mem_iff in Hlocin.
    pose (SetFacts.elements_iff (elt:=locT) locs loc) as Hina.
    inversion Hina as [Hina_to Hina_from].
    pose (Hina_to Hlocin) as Heltsin.
    assert (forall init,
              (mem) [loc ]!
              ⊑
              fold_left (λ (a : valueT) (e : locT), (mem) [e ]! ⊔ a)
              (SetInterface.elements locs) init) as Hgen.
    { elim: (SetInterface.elements locs) Heltsin => [|hd tl IH] Heltsin //;
      first by inversion Heltsin.
      move=> init.
      rewrite/fold_left.
      inversion Heltsin; subst.
      { rewrite H0. 
        rewrite -> (memory_lookup_fold_spec tl (mem[hd]!) init mem).
        apply/join_lhs.
      }
      { apply IH; done. }
    }
    apply Hgen.
  Qed.

  Lemma memory_lookup_lub :
    forall (locs:Set![locT]) (mem:memT) (vub:valueT)
           (Hub: forall loc, loc !\in locs -> le mem[loc]! vub),
      le (memory_lookup (Elt locs) mem) vub.
  Proof.
    move=> locs mem vub Hlocin.
    rewrite/memory_lookup.
    rewrite SetInterface.fold_1.
    assert (Heltfact : (forall (x:locT), InA (_eq (OrderedType:=order_OrderedType)) x (SetInterface.elements locs) -> le mem[x]! vub)).
    { move=> loc Hloc.
      apply Hlocin.
      rewrite/smem/set_setlike.
      apply SetFacts.mem_iff.
      apply SetFacts.elements_iff.
      rewrite/_eq in Hloc.
      done.
    }
    elim: (SetInterface.elements locs) {Hlocin} Heltfact
    => [|lhd ltl IH] Heltfact.
    { subst; exact:bot_prop. }
    { rewrite{1}/fold_left.
      rewrite -> (memory_lookup_fold_spec ltl (mem[lhd]!) bot mem).
      apply/join_lub.
      { apply Heltfact.
        apply InA_cons_hd; done.
      }
      { apply IH.
        move=> loc Hloc.
        apply Heltfact.
        by right.
      }
    }
  Qed.

  Global Add Parametric Morphism : memory_lookup with
      signature eq ==> eq ==> eq as memory_lookup_mor.
  Proof.
    move=> llocs rlocs Hlocs lm rm Hm.
    apply/DomType.le_antisym.
    { case: llocs Hlocs => [llocs|] Hlocs //;
      case: rlocs Hlocs => [rlocs|] Hlocs //.
      rewrite/eq/eq_/withtop_setoid/withtop_eq
      /eq/eq_/set_setoid/SetInterface.Equal in Hlocs.
      apply memory_lookup_lub; move=> loc Hlloc.
      assert (loc !\in rlocs) as Hrloc.
      { apply SetFacts.mem_iff; apply Hlocs; apply SetFacts.mem_iff.
        exact:Hlloc.
      }
      rewrite <- (memory_lookup_ub rlocs rm loc Hrloc).
      rewrite Hm; apply/DomType.eq_le/eq_refl.
    }
    { case: llocs Hlocs => [llocs|] Hlocs //;
      case: rlocs Hlocs => [rlocs|] Hlocs //.
      rewrite/eq/eq_/withtop_setoid/withtop_eq
      /eq/eq_/set_setoid/SetInterface.Equal in Hlocs.
      apply memory_lookup_lub; move=> loc Hrloc.
      assert (loc !\in llocs) as Hlloc.
      { apply SetFacts.mem_iff; apply Hlocs; apply SetFacts.mem_iff.
        exact:Hrloc.
      }
      rewrite <- (memory_lookup_ub llocs lm loc Hlloc).
      rewrite Hm; apply/DomType.eq_le/eq_refl.
    }
  Qed.

  Global Add Parametric Morphism : memory_update with
      signature eq ==> eq ==> le ==> le as memory_update_le_mor.
  Proof.
    move=> x y Hxy z w Hzw a b Hab.
    move: b Hab=> [b|] Hab; last exact: top_prop.
    move: a Hab=> [a|] Hab; last done.
    move: y Hxy=> [y|] Hxy; last exact: top_prop.
    move: x Hxy=> [x|] Hxy; last done.
    rewrite/eq/= in Hxy.
    rewrite/le/= in Hab.
    apply mem_le_spec; [done|done|move=> loc].
    rewrite ? memory_update_spec.
    rewrite <- Hxy.
    case: (_ !\in _); last by rewrite/=.
    rewrite <- Hxy.
    case: (eqb _ _)=> /=; last first.
    { apply join_le_mor; first done.
      exact/DomType.eq_le.
    }
    case: (is_strong _ _); first exact/DomType.eq_le.
    apply join_le_mor; first done.
    exact/DomType.eq_le.
  Qed.

  Global Add Parametric Morphism : memory_update with
      signature eq ==> eq ==> eq ==> eq as memory_update_mor.
  Proof.
    move=> x y Hxy z w Hzw a b Hab.
    apply DomType.le_antisym; eapply memory_update_le_mor; try done.
    exact/DomType.eq_le.
    exact/eq_sym.
    exact/eq_sym.
    exact/DomType.eq_le/eq_sym.
  Qed.
End Memories.
