Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomBridge.
Require Import DomSet.
Require Import DomMap.
Require Import DomZ.
Require Import DomItv.

Require Import Containers.SetProperties.
Require Import Containers.MapInterface.

Set Implicit Arguments.

Section Locs.
  (* Locations are (1) dynamic location, (2) global location, and (3)
  local location, with fields *)

  (* CAUTION: order of field

     The order of field is reverse of field access.  For example, A
     location, "x->a->b", accesses fields a and b, and they are
     attached to the x location as [b;a], NOT [a;b].  Be careful not
     to make inconsistencies when you translate and validate a
     program. *)

  Definition locT := ((allocT + idT + (fidT * idT)) * (list idT))%type.

  Global Instance locT_setoid : setoid locT := _.
  Global Instance locT_usual_setoid : @usual_setoid locT _ := _.
  Global Instance locT_setoidDec : @setoidDec locT _ := _.
  Global Instance locT_poset : @poset locT _ := _.
  Global Instance locT_posetDec : @posetDec locT _ _ _ := _.

  Definition lheap (a:allocT) (fs:list idT) : locT := (inl (inl a), fs).
  Definition lglobal (id:idT) (fs:list idT) : locT := (inl (inr id), fs).
  Definition llocal (fid:fidT) (id:idT) (fs:list idT) : locT := (inr (fid, id), fs).
  Definition var_loc (fid:fidT) (var:varT) (fs:list idT):=
    match var with
      | inl gvar => lglobal gvar fs
      | inr lvar => llocal fid lvar fs
    end.

  Global Add Parametric Morphism : lheap with
      signature eq ==> eq ==> eq as lheap_mor.
  Proof. done. Qed.

  Global Add Parametric Morphism : lglobal with
      signature eq ==> eq ==> eq as lglobal_mor.
  Proof. done. Qed.

  Global Add Parametric Morphism : llocal with
      signature eq ==> eq ==> eq ==> eq as llocal_mor.
  Proof. done. Qed.

  Global Add Parametric Morphism : var_loc with
      signature eq ==> eq ==> eq ==> eq as var_loc_mor.
  Proof.
    intros; rewrite/var_loc; case: x0 H0; case: y0 => //=.
  Qed.

  Definition locsT := withtop Set![locT].

  Program Definition filter_fid (fid:fidT) : \powerset locT :=
    fun loc =>
      if loc is (inr (fid', _), _)
      then ~~ (eqb fid fid')
      else true.
  Next Obligation.
    move=> [[[x|x]|x] xs] [[[y|y]|y] ys] // [Hxy Hxys]; try done.
    rewrite/fst/snd in Hxy Hxys.
    move: x y Hxy=> [xf xv] [yf yv] Hxy.
    by usual_tac Hxy; inv Hxy.
  Qed.
End Locs.

Section Osss.
  Definition osT := (itvT * itvT)%type.
  Global Instance osT_poset : @poset osT _ := pair_poset_pointwise.
  Global Instance osT_hasBot : @hasBot osT _ osT_poset :=
    pair_hasBot_pointwise.
  Global Instance osT_hasJoin : @hasJoin osT _ osT_poset :=
    pair_hasJoin_pointwise.

  Definition ossT := (osT * itvT)%type.
  Global Instance ossT_poset : @poset ossT _ := pair_poset_pointwise.
  Global Instance ossT_hasBot : @hasBot ossT _ ossT_poset :=
    pair_hasBot_pointwise (H3:=osT_hasBot).
  Global Instance ossT_hasJoin : @hasJoin ossT _ ossT_poset :=
    pair_hasJoin_pointwise (H3:=osT_hasJoin).

  Definition offset_of (oss:ossT) : itvT := oss.1.1.
  Definition size_of (oss:ossT) : itvT := oss.1.2.
  Definition stride_of (oss:ossT) : itvT := oss.2.

  Definition offset_to_ossT (o:itvT) : ossT := (o, bot, bot).
  Definition size_to_ossT (s:itvT) : ossT := (bot, s, bot).
  Definition stride_to_ossT (st:itvT) : ossT := (bot, bot, st).
End Osss.


Section Arrvs.
  Definition arrvT' := Map![allocT, ossT].
  Global Instance arrvT'_poset : @poset arrvT' _.
  Proof.
    apply (map_poset (H1:=ossT_poset)).
  Defined.
  Global Instance arrvT'_hasBot : @hasBot arrvT' _ arrvT'_poset := _.    
  Global Instance arrvT'_hasJoin : @hasJoin arrvT' _ arrvT'_poset := _.

  Definition arrvT := withtop arrvT'.
  Global Instance arrvT_poset : @poset arrvT _.
  Proof.
    apply @withtop_poset.
    apply (map_poset (H1:=ossT_poset)).
  Defined.
  Global Instance arrvT_hasBot : @hasBot arrvT _ arrvT_poset := _.
  Global Instance arrvT_hasJoin : @hasJoin arrvT _ arrvT_poset := _.

  Definition locs_from_arrvT (arrv:arrvT) : locsT :=
    match arrv with
      | Elt av =>
        Elt (fold
               (fun k v s =>
                  if ~~ eqb v bot
                  then Containers.SetInterface.add (lheap k nil) s
                  else s)
               av
               Containers.SetInterface.empty
            )
      | _ => top
    end.

  Lemma locs_from_arrvT_locs_spec_1 :
    forall av l locs
           (Hlocs: locs_from_arrvT (Elt av) = Elt locs)
           (Hin: SetInterface.In l locs),
    exists h, eq l (lheap h nil).
  Proof.
    move=> av l locs Hlocs Hin.
    rewrite/locs_from_arrvT in Hlocs; inversion Hlocs as [Hlocs'];
    clear Hlocs.
    rewrite <- Hlocs' in Hin; clear Hlocs'.
    rewrite fold_1 in Hin.
    rewrite <- fold_left_rev_right in Hin.
    elim: (rev (elements av)) Hin => [|[hdk hdv] tl IH] Hin.
    { simpl in Hin; inversion Hin. }
    { rewrite{1}/fold_right in Hin; simpl in Hin.
      case: (~~ eqb hdv bot) Hin => Hin.
      { rewrite -> (SetFacts.add_iff (elt:=locT)) in Hin.
        inversion Hin as [Hadded|Horig]; clear Hin.
        { exists hdk; by exact:eq_sym. }
        { exact: (IH Horig). }
      }
      { exact: (IH Hin). }
    }
  Qed.

  Lemma locs_from_arrvT_locs_spec_2 :
    forall av h locs alloc
           (Hllocs : locs_from_arrvT (Elt av) = Elt locs)
           (Heq : eq alloc (lheap h nil))
           (Hin : SetInterface.In alloc locs),
      ~~ eqb (coerce (av) [h]) bot = true.
  Proof.
    move=> av h locs alloc Hlocs Heq Hin.
    rewrite -> Heq in Hin; clear Heq.
    rewrite/locs_from_arrvT in Hlocs; inversion Hlocs as [Hlocs'];
    clear Hlocs.
    rewrite <- Hlocs' in Hin; clear Hlocs'.
    rewrite fold_1 in Hin.
    rewrite <- fold_left_rev_right in Hin.
    rewrite MF.elements_o.
    rewrite MF.findA_rev; last first.
    apply elements_3w.
    pose (elements_3w av) as Hnodup.
    apply NoDupA_rev in Hnodup.
    elim: {av} (rev (elements av)) Hin Hnodup
    => [|[hdk hdv] tl IH] Hin Hnodup.
    { simpl in Hin; inversion Hin. }
    { rewrite{1}/fold_right in Hin; simpl in Hin.
      case Heqb: (~~ eqb hdv bot) Hin => Hin.
      { rewrite -> (SetFacts.add_iff (elt:=locT)) in Hin.
        inversion Hin as [[Hadded _]|Horig]; clear Hin.
        { inversion Hadded; subst.
          by rewrite/=/MF.eqb is_compare_eqb eqb_refl /coerce.
        }
        { rewrite{1}/findA.
          inversion Hnodup as [|x l Hduphd Hduptl]; clear Hnodup.
          case: (MF.eqb h hdk); first done.
          exact: (IH Horig Hduptl).
        }
      }
      { inversion Hnodup as [|x l Hduphd Hduptl]; subst; clear Hnodup.
        rewrite{1}/findA.
        case Heqbh: (MF.eqb h hdk).
        { pose (IH Hin Hduptl) as Hcontr.
          rewrite/MF.eqb in Heqbh.
          rewrite is_compare_eqb in Heqbh.
          move/eqbP in Heqbh.
          elim Hduphd.
          rewrite -> Bool.negb_true_iff in Hcontr.
          rewrite/coerce in Hcontr.
          case Hfind: (findA (MF.eqb h) tl) Hcontr => [hv|] Hcontr //.
          rewrite <- MF.find_NoDup in Hfind; last done.
          clear -Heqbh Hfind.
          elim: tl Hfind => [|hd tl IH] Hfind; first by inversion Hfind.
          inversion Hfind; subst.
          { simpl in H0; inversion H0 as [Hh Hhv]; clear H0.
            apply InA_cons_hd; simpl.
            by rewrite <- Heqbh.
          }
          { by apply InA_cons_tl; apply IH. }
        }
        { exact: (IH Hin Hduptl). }
      }
    }
    { auto. }
  Qed.

  Lemma locs_from_arrvT_locs_spec_3 :
    forall av h locs
           (Hllocs: locs_from_arrvT (Elt av) = Elt locs)
           (Hnotbot: ~~ eqb (coerce (av) [h]) bot = true),
      SetInterface.In (lheap h nil) locs.
  Proof.
    move=> av h locs Hlocs Heqb.
    rewrite/locs_from_arrvT in Hlocs; inversion Hlocs as [Hlocs'];
    clear Hlocs Hlocs'.
    rewrite fold_1.
    rewrite <- fold_left_rev_right.
    rewrite -> Bool.negb_true_iff in Heqb.
    rewrite/coerce in Heqb.
    case Havh: av[h] Heqb => [avhe|] Heqb //.
    rewrite -> MF.elements_o in Havh.
    rewrite -> MF.findA_rev in Havh; last by apply elements_3w.
    elim: (rev (elements av)) Havh => [|[hdk hdv] tl IH] Havh.
    { inversion Havh. }
    { inversion Havh as [Havh']; clear Havh.
      case Heqbk: (MF.eqb h hdk) Havh' => Havh'.
      { inversion Havh'; subst; clear Havh'.
        simpl; rewrite Heqb /negb.
        rewrite (SetFacts.add_iff (elt:=locT)); left.
        rewrite/MF.eqb in Heqbk.
        rewrite is_compare_eqb in Heqbk.
        move/eqbP in Heqbk.
        rewrite Heqbk.
        exact:eq_refl.
      }
      { rewrite{1}/fold_right//=.
        case Heqbv: (~~ eqb hdv bot).
        { rewrite (SetFacts.add_iff (elt:=locT)); right.
          by apply IH.
        }
        { by apply IH. }
      }
    }
  Qed.

  Global Add Parametric Morphism : locs_from_arrvT with
    signature eq ==> eq as locs_from_arrvT_mor.
  Proof.
    move=> lav rav Hav.
    rewrite/eq/eq_/withtop_setoid/withtop_eq.
    case: lav Hav => [lav|] Hav; case: rav Hav => [rav|] Hav //.
    case Hllocs: (locs_from_arrvT (Elt lav)) => [llocs|] //.
    case Hrlocs: (locs_from_arrvT (Elt rav)) => [rlocs|] //.
    rewrite/eq/eq_/set_setoid/SetInterface.Equal.
    move=> alloc; split; move=> Hin.
    { pose (locs_from_arrvT_locs_spec_1 Hllocs Hin) as Hllspec;
      inversion Hllspec as [lh Hlh]; subst; clear Hllspec.
      exploit locs_from_arrvT_locs_spec_2;
        [exact Hllocs|by eauto|by auto|].
      move=> Hlavlh.
      rewrite Hlh.
      apply locs_from_arrvT_locs_spec_3 with (av:=rav); first done.
      rewrite/eq/eq_/withtop_setoid/withtop_eq
             /eq/eq_/map_setoid in Hav.
      pose (Hav lh) as Havlh.
      by rewrite <- Havlh.
    }
    { pose (locs_from_arrvT_locs_spec_1 Hrlocs Hin) as Hrlspec;
      inversion Hrlspec as [rh Hrh]; subst; clear Hrlspec.
      exploit locs_from_arrvT_locs_spec_2;
        [exact Hrlocs|by eauto|by auto|].
      move=> Hravrh.
      rewrite Hrh.
      apply locs_from_arrvT_locs_spec_3 with (av:=lav); first done.
      rewrite/eq/eq_/withtop_setoid/withtop_eq
             /eq/eq_/map_setoid in Hav.
      pose (Hav rh) as Havrh.
      by rewrite -> Havrh.
    }
  Qed.
End Arrvs.


Section Values.
  Definition valueT := (itvT * locsT * arrvT)%type.

  Definition itv_of (v:valueT) : itvT := v.1.1.
  Definition locs_of (v:valueT) : locsT := v.1.2.
  Definition arrv_of (v:valueT) : arrvT := v.2.

  Definition itv_to_valueT (itv:itvT) : valueT := (itv, bot, bot).
  Definition locs_to_valueT (locs:locsT) : valueT := (bot, locs, bot).
  Definition arrv_to_valueT (arrv:arrvT) : valueT := (bot, bot, arrv).

  Global Instance valueT_setoid : setoid valueT := _.
  Global Instance valueT_poset : @poset valueT _ :=
    @pair_poset_pointwise
      _ _ _ _
      (@pair_poset_pointwise
         _ _ _ _ _ _) _.
  Global Instance valueT_hasBot : @hasBot valueT _ valueT_poset := _.
  Global Instance valueT_hasTop : @hasTop valueT _ _ := _.
  Global Instance valueT_hasJoin : @hasJoin valueT _ _ := _.

  Add Parametric Morphism : itv_of with
    signature le ==> le as itv_of_le_mor.
  Proof.
    by move=> [[xa xb] xc] [[ya yb] yc] [[Ha Hb] Hc].
  Qed.
  Add Parametric Morphism : itv_of with
    signature eq ==> eq as itv_of_mor.
  Proof.
    by move=> [[xa xb] xc] [[ya yb] yc] [[Ha Hb] Hc].
  Qed.

  Add Parametric Morphism : locs_of with
    signature le ==> le as locs_of_le_mor.
  Proof.
    by move=> [[xa xb] xc] [[ya yb] yc] [[Ha Hb] Hc].
  Qed.
  Add Parametric Morphism : locs_of with
    signature eq ==> eq as locs_of_mor.
  Proof.
    by move=> [[xa xb] xc] [[ya yb] yc] [[Ha Hb] Hc].
  Qed.

  Add Parametric Morphism : arrv_of with
    signature le ==> le as arrv_of_le_mor.
  Proof.
    by move=> [[xa xb] xc] [[ya yb] yc] [[Ha Hb] Hc].
  Qed.
  Add Parametric Morphism : arrv_of with
    signature eq ==> eq as arrv_of_mor.
  Proof.
    by move=> [[xa xb] xc] [[ya yb] yc] [[Ha Hb] Hc].
  Qed.

  Add Parametric Morphism : itv_to_valueT with
    signature eq ==> eq as itv_to_valueT_mor.
  Proof. done. Qed.
 
  Add Parametric Morphism : locs_to_valueT with
    signature eq ==> eq as locs_to_valueT_mor.
  Proof. done. Qed.
 
  Add Parametric Morphism : arrv_to_valueT with
    signature eq ==> eq as arrv_to_valueT_mor.
  Proof. done. Qed.

  Section ArrayVals.
    Definition arrv_op_add (x:arrvT) (y:itvT) : arrvT :=
      match y with
        | Some _ =>
          match x with
            | Elt xe =>
              Elt (Containers.MapInterface.map
                (key:=allocT)
                (fun e => match e with
                            | (o,s,st) => (DomItv.op_add o y, s, st)
                          end)
                xe
              )
            | Top => Top arrvT'
          end
        | None => bot
      end.

    Definition arrv_op_sub (x:arrvT) (y:itvT) : arrvT :=
      match y with
        | Some _ =>
          match x with
            | Elt xe =>
              Elt (Containers.MapInterface.map
                (key:=allocT)
                (fun e => match e with
                            | (o,s,st) => (DomItv.op_sub o y, s, st)
                          end)
                xe
              )
            | Top => Top arrvT'
          end
        | None => bot
      end.

    Add Parametric Morphism : arrv_op_add with
      signature eq ==> eq ==> eq as arrv_op_add_mor.
    Proof.
      move => [x|] [y|] Hxy [a|] [b|] Hab //=.
      rewrite/eq/eq_/withtop_setoid/withtop_eq.
      intro k. do 2 rewrite MF.map_o.
      rewrite/MF.option_map.
      rewrite/eq/= in Hxy.
      move: (Hxy k)=> {Hxy}.
      move: (_[_]) (_[_])=> [xk|] [yk|] Hxy.
      { simpl in Hxy. destruct yk. destruct o. destruct xk. destruct o.
        split. split.
        destruct Hxy. destruct H. unfold fst in *. apply op_add_mor; auto.
        destruct Hxy. destruct H. unfold fst,snd in *. auto.
        destruct Hxy. destruct H. unfold fst,snd in *. auto. }
      { rewrite/coerce in Hxy *.
        destruct xk. destruct o.
        split. split.
        destruct Hxy. destruct H. unfold fst in *. 
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise in H.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise in H.
        rewrite <- H. reflexivity.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption. }
      { rewrite/coerce in Hxy *.
        destruct yk. destruct o.
        split. split.
        destruct Hxy. destruct H. unfold fst in *. 
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise in H.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise in H.
        rewrite H. reflexivity.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption. }
      { reflexivity. }
    Qed.

    Add Parametric Morphism : arrv_op_sub with
      signature eq ==> eq ==> eq as arrv_op_sub_mor.
    Proof.
      move => [x|] [y|] Hxy [a|] [b|] Hab //=.
      rewrite/eq/eq_/withtop_setoid/withtop_eq.
      intro k. do 2 rewrite MF.map_o.
      rewrite/MF.option_map.
      rewrite/eq/= in Hxy.
      move: (Hxy k)=> {Hxy}.
      move: (_[_]) (_[_])=> [xk|] [yk|] Hxy.
      { simpl in Hxy. destruct yk. destruct o. destruct xk. destruct o.
        split. split.
        destruct Hxy. destruct H. unfold fst in *. apply op_sub_mor; auto.
        destruct Hxy. destruct H. unfold fst,snd in *. auto.
        destruct Hxy. destruct H. unfold fst,snd in *. auto. }
      { rewrite/coerce in Hxy *.
        destruct xk. destruct o.
        split. split.
        destruct Hxy. destruct H. unfold fst in *. 
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise in H.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise in H.
        rewrite <- H. reflexivity.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption. }
      { rewrite/coerce in Hxy *.
        destruct yk. destruct o.
        split. split.
        destruct Hxy. destruct H. unfold fst in *. 
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise.
        rewrite/bot/bot_/ossT_hasBot/pair_hasBot_pointwise in H.
        rewrite/bot/bot_/osT_hasBot/pair_hasBot_pointwise in H.
        rewrite H. reflexivity.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption.
        destruct Hxy. destruct H. unfold fst,snd in *.
        assumption. }
      { reflexivity. }
    Qed.
  End ArrayVals.

  Program Definition vconst (z:Z) : valueT := itv_to_valueT (Some (itv'_z_z z z _)).
  Next Obligation. exact/DomType.eq_le/eq_refl. Qed.

  Definition vop1 (op:op1) (val:valueT) : valueT :=
    match op with
      | DomBasic.op_not => itv_to_valueT (DomItv.op_not (itv_of val))
      | DomBasic.op_bnot => itv_to_valueT (DomItv.op_bnot (itv_of val))
    end.

  Definition vop2 (op:op2) (x y:valueT) : valueT :=
    match op with
      | DomBasic.op_add =>
        (DomItv.op_add (itv_of x) (itv_of y),
         bot,
         join 
           (arrv_op_add (arrv_of x) (itv_of y))
           (arrv_op_add (arrv_of y) (itv_of x)))
      | DomBasic.op_sub =>
        (DomItv.op_sub (itv_of x) (itv_of y),
         bot,
         arrv_op_sub (arrv_of x) (itv_of y))
      | DomBasic.op_mul =>
        (DomItv.op_mul (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_div =>
        (DomItv.op_div (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_mod =>
        (DomItv.op_mod (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_lt =>
        (DomItv.op_lt (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_le =>
        (DomItv.op_le (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_gt =>
        (DomItv.op_gt (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_ge =>
        (DomItv.op_ge (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_eq =>
        (DomItv.op_eq (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_neq =>
        (DomItv.op_neq (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_and =>
        (DomItv.op_and (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_or =>
        (DomItv.op_or (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_band =>
        (DomItv.op_band (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_bor =>
        (DomItv.op_bor (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_bxor =>
        (DomItv.op_bxor (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_lshift =>
        (DomItv.op_lshift (itv_of x) (itv_of y), bot, bot)
      | DomBasic.op_rshift =>
        (DomItv.op_rshift (itv_of x) (itv_of y), bot, bot)
    end.

  Global Add Parametric Morphism : vop1 with
      signature eq ==> eq ==> eq as vop1_mor.
  Proof.
    move => lop rop Hop [[la lb] lc] [[ra rb] rc] [[Ha Hb] Hc].
    rewrite/= in Ha Hb Hc.
    case: lop Hop; case: rop => //; move=> Htriv;
      do ? split; solve_eq_try.
    exact: DomItv.op_not_mor.
    exact: DomItv.op_bnot_mor.
  Qed.

  Global Add Parametric Morphism : vop2 with
      signature eq ==> eq ==> eq ==> eq as vop2_mor.
  Proof.
    move=> lop rop Hop.
    move=> [[la lb] lc] [[ra rb] rc] [[Ha Hb] Hc].
    rewrite/= in Ha Hb Hc.
    move=> [[le lf] lg] [[re rf] rg] [[He Hf] Hg].
    rewrite/= in He Hf Hg.
    usual_tac Hop.
    case: rop; do ? split; solve_eq_try.
    { exact: DomItv.op_add_mor. }
    { by apply join_mor; apply: arrv_op_add_mor. }
    { exact: DomItv.op_sub_mor. }
    { exact: arrv_op_sub_mor. }
    { exact: DomItv.op_mul_mor. }
    { exact: DomItv.op_div_mor. }
    { exact: DomItv.op_mod_mor. }
    { exact: DomItv.op_lt_mor. }
    { exact: DomItv.op_le_mor. }
    { exact: DomItv.op_gt_mor. }
    { exact: DomItv.op_ge_mor. }
    { exact: DomItv.op_eq_mor. }
    { exact: DomItv.op_neq_mor. }
    { exact: DomItv.op_and_mor. }
    { exact: DomItv.op_or_mor. }
    { exact: DomItv.op_band_mor. }
    { exact: DomItv.op_bor_mor. }
    { exact: DomItv.op_bxor_mor. }
    { exact: DomItv.op_lshift_mor. }
    { exact: DomItv.op_rshift_mor. }
  Qed.
End Values.
