Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomBridge.
Require Import DomSet.
Require Import DomMap.
Require Import DomZ.
Require Import DomItv.
Require Import DomPidx.
Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.

Require Import DomBasic.
Require Import Syntax.

Require Import Containers.SetProperties.
Require Import Containers.MapInterface.

Set Implicit Arguments.

Section SemAbsBlocks.
  Variable (prog:progT) (acc:accT).

  Program Definition range_of (ct:ctype) : option itvT :=
    match ct with
      | typ_uint => Some (Some (itv'_z_z 0 4294967295 _))
      | typ_uchar => Some (Some (itv'_z_z 0 255 _))
      | typ_ushort => Some (Some (itv'_z_z 0 65535 _))
      | typ_ulong => Some (Some (itv'_z_z 0 4294967295 _))
      | typ_int => Some (Some (itv'_z_z (-2147483648) 2147483647 _))
      | typ_char => Some (Some (itv'_z_z (-128) 127 _))
      | typ_short => Some (Some (itv'_z_z (-32768) 32767 _))
      | typ_long => Some (Some (itv'_z_z (-2147483648) 2147483647 _))
      | _ => None
    end.

  Definition ctype_stride_of (ct:ctype) : option Z :=
    match ct with
      | typ_uint       => Some 4%Z
      | typ_uchar      => Some 1%Z
      | typ_ushort     => Some 2%Z
      | typ_ulong      => Some 4%Z
      | typ_ulonglong  => Some 8%Z
      | typ_int        => Some 4%Z
      | typ_char       => Some 1%Z
      | typ_short      => Some 2%Z
      | typ_long       => Some 4%Z
      | typ_longlong   => Some 8%Z
      | typ_struct _   => None
      | typ_union _    => None
      | typ_pointer _  => Some 4%Z
      | typ_array _ _  => None
      | typ_function _ => None
      | typ_void       => None
      | typ_va_list    => None
      | typ_unknown    => None
    end.

  Definition itv_cast (ct:ctype) (itv:itvT) : itvT :=
    let range := (range_of ct) in
      match range with
        | Some r => if (DomType.leb itv r) then itv else r
        | None => top
      end.

  Global Add Parametric Morphism : itv_cast with
      signature eq ==> eq ==> eq as itv_cast_mor.
  Proof.
    move=> lct rct Hct litv ritv Hitv.
    usual_tac Hct.
    rewrite/itv_cast.
    case: (range_of rct) => [elt|] //.
    apply congr_if; auto; last by apply/eq_refl.
    rewrite Hitv; done.
  Qed.

  Program Definition itv_const (z:Z) : itvT := Some (itv'_z_z z z _).
  Next Obligation. rewrite/le/=; omega. Qed.

  Definition arrv_cast (ct:ctype) (arrv:arrvT) : arrvT :=
    match arrv with
      | Elt av => (
          match ct with
            | typ_pointer pt => (
                match (ctype_stride_of pt) with
                  | Some st =>
                    let newStride := (itv_const st) in
                    Elt (Containers.MapInterface.map
                           (fun v =>
                              if (eqb v bot) then v else
                                match v with
                                  | (os,size,stride) =>
                                    ((DomItv.op_div (DomItv.op_mul os stride)
                                                    newStride),
                                     (DomItv.op_div (DomItv.op_mul size stride)
                                                    newStride),
                                     newStride)
                                end
                           )
                           av
                        )
                  | _ => arrv
                end
              )
            | _ => arrv
          end
        )
      | _ => top
    end.

  Global Add Parametric Morphism : arrv_cast with
      signature eq ==> eq ==> eq as arrv_cast_mor.
  Proof.
    move=> lct rct Hct lav rav Hav.
    usual_tac Hct.
    rewrite/arrv_cast.
    case: lav Hav => [lav'|] Hav; case: rav Hav => [rav'|] Hav //=.
    case: rct => //; move=> ct.
    case: (ctype_stride_of ct) => //; move=> z.
    dom_tac.
    rewrite/eq/eq_/withtop_setoid/withtop_eq in Hav.
    move=> loc.
    rewrite ? MF.map_o.
    rewrite/MF.option_map.
    rewrite/eq/eq_/map_setoid in Hav.
    pose (Hav loc) as Havloc.
    case: lav' [loc] Havloc => [loss|] Havloc;
    case: rav' [loc] Havloc => [ross|] Havloc //.
    { apply coerce_eq_mor.
      rewrite/eq/eq_/option_setoid/DomBasic.option_eq.
      rewrite/coerce in Havloc.
      case: loss Havloc => [[los lsize] lstride] Havloc;
      case: ross Havloc => [[ros rsize] rstride] Havloc;
      case Hlbot: (eqb (los, lsize, lstride) bot);
      case Hrbot: (eqb (ros, rsize, rstride) bot) => //=;
      move/eqbP in Hlbot; move/eqbP in Hrbot.
      { rewrite -> Hlbot in Havloc.
        elim Hrbot; apply eq_sym; done.
      }
      { rewrite -> Hrbot in Havloc.
        elim Hrbot; done.
      }
      { inversion Havloc as [Hos Hstr]; simpl in *.
        inversion Hos as [Ho Hsize]; simpl in *.
        split; last done.
        { split; simpl.
          { rewrite Hstr; rewrite Ho; exact:eq_refl. }
          { rewrite Hstr; rewrite Hsize; exact:eq_refl. }
        }
      }
    }
    { rewrite/coerce in Havloc.
      rewrite/coerce.
      case Heqbot: (eqb loss bot); first done.
      move/eqbP in Heqbot; elim Heqbot; done.
    }
    { rewrite/coerce in Havloc.
      rewrite/coerce.
      case Heqbot: (eqb ross bot); first done.
      move/eqbP in Heqbot; elim Heqbot; exact:eq_sym.
    }
  Qed.

  Definition cast (ct:ctype) (v:valueT) : valueT :=
    match v with
      | (itv, _, arrv) =>
        (itv_cast ct itv, bot, arrv_cast ct arrv)
    end.

  Global Add Parametric Morphism : cast with
      signature eq ==> eq ==> eq as cast_mor.
  Proof.
    move=> x y Hxy [[z w] a] [[c d] e] [[Hzc Hwd] Hae].
    usual_tac Hxy.
    rewrite/fst/snd in Hzc Hwd Hae.
    rewrite/cast.
    do ? split; rewrite/fst/snd; try done.
    by rewrite Hzc; reflexivity.
    by rewrite Hae; reflexivity.
  Qed.

  Program Definition append_field (f:idT) : locT ->> locT :=
    fun loc => (loc.1, f::loc.2).
  Next Obligation.
    move=> [x xf] [y yf] [Hxy Hxyf].
    by rewrite/fst/snd in Hxy Hxyf *.
  Qed.
  
  Fixpoint sem_expr (fid:fidT) (e:exprT) (mem:memT): valueT :=
    match e with
      | expr_const lit =>
        match lit with
          | lit_int z => vconst z
          | lit_int_top => (top,bot,bot)
          | lit_top => top
          | lit_bottom => bot
        end
      | expr_op2 op lhs rhs =>
        vop2 op (sem_expr fid lhs mem) (sem_expr fid rhs mem)
      | expr_op1 op e' => vop1 op (sem_expr fid e' mem)
      | expr_value_of l => memory_lookup (sem_expr_lv fid l mem) mem
      | expr_addr_of l => (bot, sem_expr_lv fid l mem, bot)
      | expr_cast ct ce => cast ct (sem_expr fid ce mem)
      | expr_size_of _ => (top, bot, bot)
    end
  with sem_expr_lv (fid:fidT) (l:lv) (mem:memT) : locsT :=
    match l with
      | lv_id x => (Elt (Containers.SetInterface.singleton (var_loc fid x nil)))
      | lv_deref e =>
        let v := sem_expr fid e mem in
        join (locs_of v) (locs_from_arrvT (arrv_of v))
      | lv_array e _ =>
        let v := sem_expr fid e mem in
        (locs_from_arrvT (arrv_of v))
      | lv_struct l pos =>
        let locs := sem_expr_lv fid l mem in
        if locs is Elt locs'
        then Elt (DomSet.map
                    (append_field pos)
                    locs')
        else top
    end.

  Lemma sem_expr_eq_proper :
    forall {lfid rfid} (Hfid: eq lfid rfid)
           {le re} (He: eq le re)
           {lm rm} (Hm: eq lm rm),
      eq (sem_expr lfid le lm)
         (sem_expr rfid re rm)
  with sem_expr_lv_eq_proper :
    forall {lfid rfid} (Hfid: eq lfid rfid)
           {llv rlv} (Hlv: eq llv rlv)
           {lm rm} (Hm: eq lm rm),
      eq (sem_expr_lv lfid llv lm)
         (sem_expr_lv rfid rlv rm).
  Proof.
    { intros; usual_tac Hfid; usual_tac He.
      elim: re => [lit|op le IHle re IHre|op e IHe|l|l|ct e IHe|e IHe] //=.
      { by case:lit. }
      { rewrite IHle; rewrite IHre; apply eq_refl. }
      { rewrite IHe; apply eq_refl. }
      { apply/memory_lookup_mor; last done.
        apply/sem_expr_lv_eq_proper => //.
      }
      { rewrite/eq/eq_/valueT_setoid/pair_setoid/pair_eq//=; split; last done.
        split; rewrite/fst/snd; first done.
        by apply sem_expr_lv_eq_proper.
      }
      { rewrite/cast.
        case: (sem_expr rfid e lm) IHe=> [[litv llocs] larrv] IHe.
        case: (sem_expr rfid e rm) IHe => [[ritv rlocs] rarrv] IHe.
        move: IHe=> [[Hitv Hlocs] Harrv].
        rewrite/fst/snd in Hitv Hlocs Harrv.
        do ? split; rewrite/fst/snd; try done.
        by apply itv_cast_mor.
        by apply arrv_cast_mor.
      }
    }
    { intros; usual_tac Hfid; usual_tac Hlv.
      elim: rlv => [x|e|arr idx|strct Hlocs pos] //=.
      { apply join_mor.
        apply sem_expr_eq_proper; done.
        apply/locs_from_arrvT_mor/arrv_of_mor/sem_expr_eq_proper; done.
      }
      { apply/locs_from_arrvT_mor/arrv_of_mor/sem_expr_eq_proper; done. }
      { move: (sem_expr_lv _ _ _) (sem_expr_lv _ _ _) Hlocs=> [llocs|] [rlocs|] Hlocs //=.
        dom_tac.
        apply map_mor; by [reflexivity | apply Hlocs].
      }
    }
  Qed.

  Global Add Parametric Morphism : sem_expr with
      signature eq ==> eq ==> eq ==> eq as sem_expr_mor.
  Proof. intros; exact:sem_expr_eq_proper. Qed.

  Global Add Parametric Morphism : sem_expr_lv with
      signature eq ==> eq ==> eq ==> eq  as sem_expr_lv_mor.
  Proof. intros; exact:sem_expr_lv_eq_proper. Qed.

  Fixpoint sem_exprs (fid:fidT) (exprs:list exprT) (mem:memT): list valueT :=
    match exprs with
      | nil => nil
      | istr::exprs' => (sem_expr fid istr mem)::(sem_exprs fid exprs' mem)
    end.

  Global Add Parametric Morphism : sem_exprs with
      signature eq ==> eq ==> eq ==> eq as sem_exprs_mor.
  Proof.
    move=> lf rf Hf; usual_tac Hf.
    move=> les res Hes.
    apply usual_eqP' in Hes; subst.
    elim: res; first by move=> //=.
    move=> e es IH lm rm Hm; split; last exact/IH.
    by rewrite -> Hm; apply/eq_refl.
  Qed.

  Lemma sem_exprs_simpl (fid:fidT) (expr:exprT) (exprs:list exprT) (mem:memT) :
    sem_exprs fid (expr::exprs) mem =
    (sem_expr fid expr mem) :: (sem_exprs fid exprs mem).
  Proof. done. Qed.

  Section Prunings.

    Lemma Z_min_le : forall z x y (Hzx : z ⊑ x) (Hzy : z ⊑ y),
                  z ⊑ Z.min y x.
    Proof. intros; apply Z_min_glb_iff; auto. Qed.

    Definition itv_restrict_ub (x:itvT) (ub:Z) :=
      match x with
        | Some x' => 
          match x' with
            | itv'_top => Some (itv'_inf_z ub)
            | itv'_inf_z ux => Some (itv'_inf_z (Zmin ux ub))
            | itv'_z_inf lx =>
              match Z_le_dec lx ub with
                | left lxub => Some (itv'_z_z lx ub lxub)
                | right _ => None
              end
            | itv'_z_z lx ux Hluxp =>
              match Z_le_dec lx ub with
                | left Hlubp =>
                  Some (itv'_z_z lx (Zmin ux ub) (Z_min_le Hlubp Hluxp))
                | right _ => None
              end
          end
        | None => None
      end.

    Global Add Parametric Morphism : itv_restrict_ub with
        signature eq ==> eq ==> eq as itv_restrict_ub_mor.
    Proof.
      move=> litv ritv Hitv lz rz Hz.
      usual_tac Hz.
      rewrite/itv_restrict_ub.
      case: litv Hitv => [[|lrz|llz|llz lrz Hllrz]|] Hitv //;
      case: ritv Hitv => [[|rrz|rlz|rlz rrz Hrlrz]|] Hitv //.
      { inversion Hitv; subst; exact:eq_refl. }
      { inversion Hitv; subst; exact:eq_refl. }
      { inversion Hitv as [Hitvl Hitvr].
        usual_tac Hitvl; usual_tac Hitvr.
        case: (Z_le_dec rlz rz) => Hllzrz //.
      }
    Qed.

    Lemma Z_max_le : forall x y z (Hxz: x ⊑ z) (Hyz: y ⊑ z),
                       Z.max x y ⊑ z.
    Proof. intros; apply Z_max_lub_iff; auto. Qed.

    Definition itv_restrict_lb (x:itvT) (lb:Z) :=
      match x with
        | Some x' => 
          match x' with
            | itv'_top => Some (itv'_z_inf lb)
            | itv'_inf_z ux => 
              match Z_le_dec lb ux with
                | left lbux => Some (itv'_z_z lb ux lbux)
                | right _ => None
              end
            | itv'_z_inf lx => Some (itv'_z_inf (Zmax lx lb))
            | itv'_z_z lx ux Hluxp =>
              match Z_le_dec lb ux with
                | left Hlubp =>
                  Some (itv'_z_z (Zmax lx lb) ux (Z_max_le Hluxp Hlubp))
                | right _ => None
              end
          end
        | None => None
      end.

    Global Add Parametric Morphism : itv_restrict_lb with
        signature eq ==> eq ==> eq as itv_restrict_lb_mor.
    Proof.
      move=> litv ritv Hitv lz rz Hz.
      usual_tac Hz.
      rewrite/itv_restrict_lb.
      case: litv Hitv => [[|lrz|llz|llz lrz Hllrz]|] Hitv //;
      case: ritv Hitv => [[|rrz|rlz|rlz rrz Hrlrz]|] Hitv //.
      { inversion Hitv; subst; exact:eq_refl. }
      { inversion Hitv; subst; exact:eq_refl. }
      { inversion Hitv as [Hitvl Hitvr].
        usual_tac Hitvl; usual_tac Hitvr.
        case: (Z_le_dec rz rrz) => Hllzrz //.
      }
    Qed.

    Definition adjust_inter_lt (x:itvT) (y:itvT) := 
      match y with
        | Some y' =>
          match y' with
            | itv'_inf_z uy => itv_restrict_ub x (uy-1)
            | itv'_z_z _ uy _ => itv_restrict_ub x (uy-1)
            | _ => x
          end
        | None => None
      end.

    Global Add Parametric Morphism : adjust_inter_lt with
        signature eq ==> eq ==> eq as adjust_inter_lt_mor.
    Proof.
      move=> litv ritv Hitv laitv raitv Haitv.
      rewrite/adjust_inter_lt.
      case: laitv Haitv => [[|larz|lalz|lalz larz Hlalrz]|] Haitv //;
      case: raitv Haitv => [[|rarz|ralz|ralz rarz Hralrz]|] Haitv //.
      { inversion Haitv; subst.
        rewrite Hitv; exact:eq_refl.
      }
      { inversion Haitv as [Haitvl Haitvr].
        rewrite Haitvr; rewrite Hitv; exact:eq_refl.
      }
    Qed.

    Definition adjust_inter_le (x:itvT) (y:itvT) := 
      match y with
        | Some y' =>
          match y' with
            | itv'_inf_z uy => itv_restrict_ub x uy
            | itv'_z_z _ uy _ => itv_restrict_ub x uy
            | _ => x
          end
        | None => None
      end.

    Global Add Parametric Morphism : adjust_inter_le with
        signature eq ==> eq ==> eq as adjust_inter_le_mor.
    Proof.
      move=> litv ritv Hitv laitv raitv Haitv.
      rewrite/adjust_inter_le.
      case: laitv Haitv => [[|larz|lalz|lalz larz Hlalrz]|] Haitv //;
      case: raitv Haitv => [[|rarz|ralz|ralz rarz Hralrz]|] Haitv //.
      { inversion Haitv; subst; rewrite Hitv; exact:eq_refl. }
      { inversion Haitv as [Haitvl Haitvr].
        rewrite Haitvr; rewrite Hitv; exact:eq_refl.
      }
    Qed.

    Definition adjust_inter_gt (x:itvT) (y:itvT) := 
      match y with
        | Some y' =>
          match y' with
            | itv'_z_inf ly => itv_restrict_lb x (ly+1)
            | itv'_z_z ly _ _ => itv_restrict_lb x (ly+1)
            | _ => x
          end
        | None => None
      end.

    Global Add Parametric Morphism : adjust_inter_gt with
        signature eq ==> eq ==> eq as adjust_inter_gt_mor.
    Proof.
      move=> litv ritv Hitv laitv raitv Haitv.
      rewrite/adjust_inter_gt.
      case: laitv Haitv => [[|larz|lalz|lalz larz Hlalrz]|] Haitv //;
      case: raitv Haitv => [[|rarz|ralz|ralz rarz Hralrz]|] Haitv //.
      { inversion Haitv; subst; rewrite Hitv; exact:eq_refl. }
      { inversion Haitv as [Haitvl Haitvr].
        rewrite Haitvl; rewrite Hitv; exact:eq_refl.
      }
    Qed.

    Definition adjust_inter_ge (x:itvT) (y:itvT) := 
      match y with
        | Some y' =>
          match y' with
            | itv'_z_inf ly => itv_restrict_lb x ly
            | itv'_z_z ly _ _ => itv_restrict_lb x ly
            | _ => x
          end
        | None => None
      end.

    Global Add Parametric Morphism : adjust_inter_ge with
        signature eq ==> eq ==> eq as adjust_inter_ge_mor.
    Proof.
      move=> litv ritv Hitv laitv raitv Haitv.
      rewrite/adjust_inter_ge.
      case: laitv Haitv => [[|larz|lalz|lalz larz Hlalrz]|] Haitv //;
      case: raitv Haitv => [[|rarz|ralz|ralz rarz Hralrz]|] Haitv //.
      { inversion Haitv; subst; rewrite Hitv. exact:eq_refl. }
      { inversion Haitv as [Haitvl Haitvr].
        rewrite Haitvl; rewrite Hitv; exact:eq_refl.
      }
    Qed.

    Definition adjust_inter_eq (x:itvT) (y:itvT) := 
      adjust_inter_le (adjust_inter_ge x y) y.

    Global Add Parametric Morphism : adjust_inter_eq with
        signature eq ==> eq ==> eq as adjust_inter_eq_mor.
    Proof.
      move=> litv ritv Hitv laitv raitv Haitv.
      rewrite/adjust_inter_eq.
      rewrite Hitv; rewrite Haitv; exact:eq_refl.
    Qed.

    Definition adjust_inter_neq (x:itvT) (y:itvT) :=
      join (adjust_inter_lt x y) (adjust_inter_gt x y).

    Global Add Parametric Morphism : adjust_inter_neq with
        signature eq ==> eq ==> eq as adjust_inter_neq_mor.
    Proof.
      move=> litv ritv Hitv laitv raitv Haitv.
      rewrite/adjust_inter_neq.
      rewrite Hitv; rewrite Haitv; exact:eq_refl.
    Qed.

    Definition adjust_inter (x:itvT) (o:op2) (y:itvT) :=
      match o with
        | op_lt => adjust_inter_lt x y
        | op_le => adjust_inter_le x y
        | op_gt => adjust_inter_gt x y
        | op_ge => adjust_inter_ge x y
        | op_eq => adjust_inter_eq x y
        | op_neq => adjust_inter_neq x y
        | _ => x
      end.

    Global Add Parametric Morphism : adjust_inter with
        signature eq ==> eq ==> eq ==> eq as adjust_inter_mor.
    Proof.
      move=> litv ritv Hitv lop rop Hop laitv raitv Haitv.
      usual_tac Hop.
      rewrite/adjust_inter.
      case: rop => //; rewrite Hitv; rewrite Haitv; exact:eq_refl.
    Qed.

    Definition adjust_loc_eq (x:locsT) (y:locsT) :=
      match y with
        | Elt y' =>
          match x with
            | Elt x' => Elt (SetInterface.inter x' y')
            | Top => y
          end
        | Top => x
      end.

    Global Add Parametric Morphism : adjust_loc_eq with
        signature eq ==> eq ==> eq as adjust_loc_eq_mor.
    Proof.
      move=> lxlocs rxlocs Hxlocs lylocs rylocs Hylocs.
      rewrite/adjust_loc_eq.
      case: lylocs Hylocs => [lylocs|] Hylocs;
      case: rylocs Hylocs => [rylocs|] Hylocs;
      case: lxlocs Hxlocs => [lxlocs|] Hxlocs;
      case: rxlocs Hxlocs => [rxlocs|] Hxlocs //.
      apply SetFacts.inter_m; done.
    Qed.

    Definition adjust_loc (x:locsT) (o:op2) (y:locsT) :=
      match o with
        | op_eq => adjust_loc_eq x y
        | _ => x
      end.

    Global Add Parametric Morphism : adjust_loc with
        signature eq ==> eq ==> eq ==> eq as adjust_loc_mor.
    Proof.
      move=> llocs rlocs Hlocs lop rop Hop lalocs ralocs Halocs.
      usual_tac Hop.
      case: rop => //=.
      apply adjust_loc_eq_mor; done.
    Qed.

    Definition adjust_val (x:valueT) (o:op2) (y:valueT) :=
      match x,y with
        | (ix,lx,rx),(iy,ly,ry) => 
          (adjust_inter ix o iy, adjust_loc lx o ly, rx)
      end.

    Global Add Parametric Morphism : adjust_val with
        signature eq ==> eq ==> eq ==> eq as adjust_val_mor.
    Proof.
      move=> lv rv Hv lop rop Hop lav rav Hav.
      usual_tac Hop.
      rewrite/adjust_val.
      case: lav Hav => [[litv llocs] lav] Hav;
      case: rav Hav => [[ritv rlocs] rav] Hav;
      case: lv Hv => [[lvitv lvlocs] lvav] Hv;
      case: rv Hv => [[rvitv rvlocs] rvav] Hv //.
      move: Hav=> [[Hitv Hlocs] Hav].
      rewrite/fst/snd in Hitv Hlocs Hav.
      move: Hv=> [[Hvitv Hvlocs] Hvav].
      rewrite/fst/snd in Hvitv Hvlocs Hvav.
      split; last done.
      split; rewrite/fst/snd.
      { by apply adjust_inter_mor. }
      { by apply adjust_loc_mor. }
    Qed.      

    Definition adjust (a:locsT) (b:op2) (v:valueT) (m:memT) :=
      let av := memory_lookup a m in
        memory_update acc a (adjust_val av b v) m.

    Global Add Parametric Morphism : adjust with
        signature eq ==> eq ==> eq ==> eq ==> eq as adjust_mor.
    Proof.
      move=> llocs rlocs Hlocs lop rop Hop lv rv Hv lm rm Hm.
      usual_tac Hop.
      rewrite/adjust.
      apply memory_update_mor; try done.
      apply adjust_val_mor; try done.
      rewrite Hm; rewrite Hlocs; exact:eq_refl.
    Qed.

    Definition prune_bop fid (e:exprT) (m:memT) : memT :=
      match e with
        | expr_op2 o (expr_value_of (lv_id x)) e2 => 
          let locx := var_loc fid x nil in
            if is_strong acc locx
              then adjust (sem_expr_lv fid (lv_id x) m) o (sem_expr fid e2 m) m
              else m
        | _ => m
      end.

    Global Add Parametric Morphism : prune_bop with
        signature eq ==> eq ==> eq ==> eq as prune_bop_mor.
    Proof.
      move=> lfid rfid Hfid le re He lm rm Hm.
      usual_tac Hfid; usual_tac He.
      rewrite/prune_bop.
      case: re => //; intros.
      case: lhs => //; intros.
      case: l => //; intros.
      { case:(is_strong acc (var_loc rfid x nil)); auto.
        apply/adjust_mor; try exact:eq_refl; try done.
        { rewrite Hm; exact:eq_refl. }
      } 
    Qed.

    Fixpoint remove_uop e :=
      match e with 
        | expr_op1 op_not (expr_op1 op_not e) => remove_uop e
        | expr_op1 op_not (expr_op2 op_lt e1 e2) => 
          Some (expr_op2 op_ge e1 e2)
        | expr_op1 op_not (expr_op2 op_le e1 e2) => 
          Some (expr_op2 op_gt e1 e2)
        | expr_op1 op_not (expr_op2 op_gt e1 e2) => 
          Some (expr_op2 op_le e1 e2)
        | expr_op1 op_not (expr_op2 op_ge e1 e2) => 
          Some (expr_op2 op_lt e1 e2)
        | expr_op1 op_not (expr_op2 op_eq e1 e2) => 
          Some (expr_op2 op_neq e1 e2)
        | expr_op1 op_not (expr_op2 op_neq e1 e2) => 
          Some (expr_op2 op_eq e1 e2)
        | expr_op2 op_lt e1 e2 => Some (expr_op2 op_lt e1 e2)
        | expr_op2 op_le e1 e2 => Some (expr_op2 op_le e1 e2)
        | expr_op2 op_gt e1 e2 => Some (expr_op2 op_gt e1 e2)
        | expr_op2 op_ge e1 e2 => Some (expr_op2 op_ge e1 e2)
        | expr_op2 op_eq e1 e2 => Some (expr_op2 op_eq e1 e2)
        | expr_op2 op_neq e1 e2 => Some (expr_op2 op_neq e1 e2)
        | _ => None
      end.

    Program Definition memory_prune fid (e:exprT) (m:memT) : memT := 
      let v := sem_expr fid e m in
      if DomType.leb (itv_of v) (Some (itv'_z_z 0 0 _))
      then bot
      else match remove_uop e with
               | Some e' => prune_bop fid e' m
               | None => m
             end.

    Global Add Parametric Morphism : memory_prune with
        signature eq ==> eq ==> eq ==> eq as memory_prune_mor.
    Proof.
      move=> lfid rfid Hfid le re He lm rm Hm.
      usual_tac Hfid.
      usual_tac He.
      rewrite/memory_prune.

      case Hleqz: (DomType.leb _ _);
      case Hreqz: (DomType.leb _ _);
      first exact:eq_refl;
      move/lebP in Hleqz; move/lebP in Hreqz.
      { elim Hreqz.
        assert (eq (itv_of (sem_expr rfid re rm))
                   (itv_of (sem_expr rfid re lm))) as Hbridge.
        { apply/itv_of_mor/sem_expr_mor; try exact:eq_refl;
          last by exact:eq_sym.
        }
        rewrite -> Hbridge; done.
      }
      { elim Hleqz.
        assert (eq (itv_of (sem_expr rfid re rm))
                   (itv_of (sem_expr rfid re lm))) as Hbridge.
        { apply/itv_of_mor/sem_expr_mor; try exact:eq_refl;
          last by exact:eq_sym.
        }
        rewrite <- Hbridge; done.
      }
      {
        case: (remove_uop re); last done.
        move=> e; rewrite Hm; exact:eq_refl.
      }
    Qed.
  End Prunings.

  Lemma elements_singleton (loc:locT) : SetInterface.elements (ssingleton loc) = loc::nil.
  Proof. reflexivity. Qed.

  Definition libcall (fid:fidT) (site:allocT) (ret:option lv) (lid:lcall)
             (args:list exprT) (mem:memT) : memT :=
    match lid with
      | airac_print => mem
      | airac_dump => mem
      | rand =>
        match ret with
          | Some retlv =>
            let retlocs := sem_expr_lv fid retlv mem in
            let rand := itv_to_valueT (Some (itv'_z_inf 0)) in
            memory_update acc retlocs rand mem
          | None => mem
        end
      | strlen =>
        match ret with
          | Some retlv =>
            let retlocs := sem_expr_lv fid retlv mem in
            let length := itv_to_valueT (Some (itv'_z_inf 0)) in
            memory_update acc retlocs length mem
          | None => mem
        end
      | strcpy =>
        let dstLocs := (sem_expr fid (hd (expr_const lit_bot) args) mem) in
        let dst := locs_of dstLocs in
        let srcLocs := (sem_expr fid (hd (expr_const lit_bot) (tl args)) mem) in
        let src := locs_of srcLocs in
        let srcValue := memory_lookup src mem in
        let newMem := memory_update acc dst srcValue mem in
        match ret with
          | Some retlv =>
            let retlocs := sem_expr_lv fid retlv newMem in
            memory_update acc retlocs (locs_to_valueT dst) newMem
          | None => newMem
        end
      | scanf => 
        let addrs := List.map
                       (fun e => (locs_of (sem_expr fid e mem)))
                       args in
        fold_right
          (fun locs m =>
             memory_update acc locs (top,bot,bot) m)
          mem addrs
      | errno_location =>
        match ret with
          | Some retlv =>
            let retlocs := sem_expr_lv fid retlv mem in
            let elt := (itv_const 0, itv_const 1, itv_const 4) in
            let singleton := Containers.MapInterface.add site elt bot in
            memory_update acc retlocs (arrv_to_valueT (Elt singleton)) mem
          | None => mem
        end
      | builtin_alloca =>
        match ret with
          | Some retlv =>
            let retlocs := sem_expr_lv fid retlv mem in
            let allocSize :=
                itv_of (sem_expr fid (hd (expr_const lit_bot) args) mem) in
            let elt := (itv_const 0, allocSize, itv_const 1) in
            let singleton := Containers.MapInterface.add site elt bot in
            memory_update acc retlocs (arrv_to_valueT (Elt singleton)) mem
          | None => mem
        end
      | realloc => mem
      | builtin_choose_expr =>
        match ret with
          | Some retlv =>
            let retlocs := sem_expr_lv fid retlv mem in
            let cond :=
                itv_of (sem_expr fid (hd (expr_const lit_bot) args) mem) in
            let v1 := sem_expr fid (hd (expr_const lit_bot) (tl args)) mem in
            let v2 := sem_expr fid (hd (expr_const lit_bot)
                                       (tl (tl args))) mem in
            if (eqb cond (itv_const 0)) then
              memory_update acc retlocs v2 mem
            else if (DomType.leb (itv_const 0) cond) then
              memory_update acc retlocs (join v1 v2) mem
            else
              memory_update acc retlocs v1 mem
          | None => mem
        end
      | unknown => mem
    end.

  Global Add Parametric Morphism : libcall with
      signature eq ==> eq ==> eq ==> eq ==> eq ==> eq ==> eq as libcall_mor.
  Proof.
    move=> lfid rfid Hfid; usual_tac Hfid.
    move=> lalc ralc Halc; usual_tac Halc.
    move=> lv rv Hv llc rlc Hlc; usual_tac Hlc.
    move=> le re He; usual_tac He.
    move=> lm rm Hm.
    case: rlc => //=.
    { case: lv Hv => [lv'|] Hv //; case: rv Hv => [rv'|] Hv //.
      inversion Hv; subst; rewrite Hm; exact:eq_refl.
    }
    { case: lv Hv => [lv'|] Hv //; case: rv Hv => [rv'|] Hv //.
      inversion Hv; subst; rewrite Hm; exact:eq_refl.
    }
    { case: lv Hv => [lv'|] Hv //; case: rv Hv => [rv'|] Hv //.
      apply memory_update_mor; inversion Hv; subst.
      { apply sem_expr_lv_mor; try done.
        apply memory_update_mor; try done.
        { apply locs_of_mor; rewrite Hm; exact:eq_refl. }
        { apply memory_lookup_mor; last done.
          apply locs_of_mor; rewrite Hm; exact:eq_refl.
        }
      }
      { apply/locs_to_valueT_mor/locs_of_mor/sem_expr_mor; try done. }
      { apply/memory_update_mor; try done.
        { apply/locs_of_mor/sem_expr_mor; try done. }
        { apply/memory_lookup_mor; last done.
          apply/locs_of_mor/sem_expr_mor; try done.
        }
      }
      { apply/memory_update_mor; try done.
        { apply/locs_of_mor/sem_expr_mor; try done. }
        { apply/memory_lookup_mor; last done.
          apply/locs_of_mor/sem_expr_mor; try done.
        }
      }
    }
    { elim: re lm rm Hm => [|elt tl IHe] lm rm Hm //.
      simpl. apply memory_update_mor; try done.
      apply/locs_of_mor/sem_expr_mor; try done.
      simpl; apply IHe; auto.
    }
    { case: lv Hv => [lv'|] Hv //; case: rv Hv => [rv'|] Hv //.
      apply/memory_update_mor; try done.
      { apply/sem_expr_lv_mor; done. }
      { apply/arrv_to_valueT_mor; exact:eq_refl. }
    }
    { case: lv Hv => [lv'|] Hv //; case: rv Hv => [rv'|] Hv //.
      apply/memory_update_mor; try done.
      { apply/sem_expr_lv_mor; try done. }
      { apply/arrv_to_valueT_mor.
        dom_tac.
        rewrite/eq/eq_/map_setoid.
        move=> key.
        apply coerce_eq_mor.
        rewrite 2! MF.add_o.
        case Hkey: (ralc == key); last done.
        rewrite/eq/eq_/option_setoid/DomBasic.option_eq.
        rewrite{1}/eq/eq_/pair_setoid/pair_eq//=.
        split; last done.
        rewrite{1}/eq/eq_//=.
        split; first done.
        apply/itv_of_mor/sem_expr_mor; try done.
      }
    }
    { case: lv Hv => [lv'|] Hv //; case: rv Hv => [rv'|] Hv //.
      inversion Hv; subst.
      case Hleq: (eqb (itv_of (sem_expr rfid (hd (expr_const lit_bot) re) lm))
                      (itv_const 0)) => //;
      case Hreq: (eqb (itv_of (sem_expr rfid (hd (expr_const lit_bot) re) rm))
                      (itv_const 0)) => //.
      { apply/memory_update_mor; try done.
        { apply/sem_expr_lv_mor; try done. }
        { apply/sem_expr_mor; try done. }
      }
      { move/eqbP in Hleq; move/eqbP in Hreq.
        elim Hreq.
        rewrite <- Hleq.
        apply/itv_of_mor/sem_expr_mor; try done.
        exact:eq_sym.
      }
      { move/eqbP in Hleq; move/eqbP in Hreq.
        elim Hleq.
        rewrite <- Hreq.
        apply/itv_of_mor/sem_expr_mor; try done.
      }
      { case Hlle: (DomType.leb
                      (itv_const 0)
                      (itv_of (sem_expr rfid (hd (expr_const lit_bot) re) lm)));
        case Hrle: (DomType.leb
                      (itv_const 0)
                      (itv_of (sem_expr rfid (hd (expr_const lit_bot) re) rm))).
        { apply/memory_update_mor; try done.
          { apply/sem_expr_lv_mor; try done. }
          { apply/join_mor.
            { apply/sem_expr_mor; try done. }
            { apply/sem_expr_mor; try done. }
          }
        }
        { move/lebP in Hlle; move/lebP in Hrle.
          elim Hrle.
          rewrite -> Hlle.
          apply DomType.eq_le.
          apply/itv_of_mor/sem_expr_mor; try done.
        }
        { move/lebP in Hlle; move/lebP in Hrle.
          elim Hlle.
          rewrite -> Hrle.
          apply DomType.eq_le.
          apply/itv_of_mor/sem_expr_mor; try done.
          exact:eq_sym.
        }
        { apply/memory_update_mor; try done.
          { apply/sem_expr_lv_mor; try done. }
          { apply/sem_expr_mor; try done. }
        }
      }
    }
  Qed.
  
  Definition sem_instr (fid:fidT) (istr:instrT) (mem:memT): memT :=
    match istr with
      | instr_assign x e =>
        let locs := sem_expr_lv fid x mem in
        memory_update acc locs (sem_expr fid e mem) mem
      | instr_alloc l a site =>
        let locs := sem_expr_lv fid l mem in
        match a with
          | alloc_arr ve ste =>
            let val := sem_expr fid ve mem in
            let st := sem_expr fid ste mem in
            let oss := (itv_const 0, itv_of val, itv_of st) in
            memory_update
              acc locs
              (itv_const 0,
               bot,
               (Elt bot[site <- oss]!))
              mem
        end
      | instr_assume tst => memory_prune fid tst mem
      | instr_libcall ret lid args site => libcall fid site ret lid args mem
      | instr_forget aloc => mremove aloc mem
      | instr_skip => mem
    end.

  Global Add Parametric Morphism : sem_instr with
      signature eq ==> eq ==> eq ==> eq as sem_instr_mor.
  Proof.
    move=> lfid rfid Hfid linstr rinstr Hinstr lm rm Hm.
    usual_tac Hfid; usual_tac Hinstr.
    case: rinstr => [var e|l alc site|test|ol lc le|l|] //.
    { simpl; eapply memory_update_mor; auto;
      by [exact:sem_expr_lv_mor|exact:sem_expr_mor].
    }
    { case: alc=> [ve ste].
      eapply memory_update_mor; eauto.
      { rewrite Hm; exact:eq_refl. }
      { split; first done.
        rewrite/fst/snd.
        dom_tac.
        rewrite/eq/eq_/map_setoid.
        move=> key.
        rewrite 2! MF.add_o.
        apply coerce_eq_mor.
        case: (site == key); last by exact:eq_refl.
        rewrite/eq/eq_/option_setoid/DomBasic.option_eq.
        rewrite/eq/eq_/pair_setoid/pair_eq//=.
               split.
        { rewrite{1}/eq/eq_.
          split; simpl; first by exact:eq_refl.
          apply itv_of_mor; rewrite Hm; exact:eq_refl.
        }
        { apply itv_of_mor; rewrite Hm; exact:eq_refl. }
      }
    }
    { simpl; rewrite Hm; exact:eq_refl. }
    { move=> site; simpl; rewrite Hm; exact:eq_refl. }
    { unfold sem_instr.
      apply mremove_mor; last done.
      by apply eq_refl.
    }
  Qed.

  Fixpoint sem_instrs (fid:fidT) (instrs:list instrT) (mem:memT): memT :=
    match instrs with
      | nil => mem
      | istr::instrs' => sem_instrs fid instrs' (sem_instr fid istr mem)
    end.

  Global Add Parametric Morphism : sem_instrs with
      signature eq ==> eq ==> eq ==> eq as sem_instrs_mor.
  Proof.
    move=> lf rf Hf listrs ristrs Histrs.
    usual_tac Hf; usual_tac Histrs.
    elim: ristrs; first by done.
    move=> i ilst IH lm rm Hm.
    apply/IH.
    by rewrite -> Hm; apply/eq_refl.
  Qed.

  Fixpoint sem_bind_params (fid:fidT) (args:list valueT)
           (params:list idT) (mem:memT): memT :=
    match args, params with
      | arg::args', param::params' =>
        sem_bind_params fid args' params'
                        (memory_update acc (Elt (ssingleton (var_loc fid (local param) nil))) arg mem)
      | _, _ => mem
    end.

  Global Add Parametric Morphism : sem_bind_params with
      signature eq ==> eq ==> eq ==> eq ==> eq as sem_bind_params_mor.
  Proof.
    move=> lfid rfid Hfid llv rlv Hlv lli rli Hli lm rm Hm.
    usual_tac Hfid; usual_tac Hli.
    elim: llv rlv Hlv rli lm rm Hm =>
    [|lvhd lvtl IHllv] rlv Hlv rli lm rm Hm //.
    { case: rlv Hlv => [|rvhd rvtl] Hlv //. }
    { case: rlv Hlv => [|rvhd rvtl] Hlv //.
      case: rli IHllv => [|ihd itl] IHllv //.
      simpl.
      apply IHllv.
      { inversion Hlv; done. }
      { apply memory_update_mor; inversion Hlv; done. }
    }
  Qed.

  Definition sem_bind_ret (fid:fidT) (result:valueT)
    (retto:option varT) (mem:memT): memT :=
    if retto is Some retto'
    then memory_update acc (Elt (ssingleton (var_loc fid retto' nil))) result mem
    else mem.

  Global Add Parametric Morphism : sem_bind_ret with
      signature eq ==> eq ==> eq ==> eq ==> eq as sem_bind_ret_mor.
  Proof.
    move=> lf rf Hf; usual_tac Hf.
    move=> lv rv Hv lvar rvar Hvar; usual_tac Hvar.
    move=> lm rm Hm.
    rewrite/sem_bind_ret.
    move: rvar=> [rvar|]; last done.
    by apply memory_update_mor.
  Qed.

  Lemma sem_bind_params_simpl : forall (fid:fidT) (arg:valueT) 
    (argvals:list valueT) (param:idT) (params:list idT) (mem:memT),
    sem_bind_params fid (arg :: argvals) (param :: params) mem =
    sem_bind_params fid argvals params
    (memory_update acc (Elt (ssingleton (var_loc fid (local param) nil))) arg mem).
  Proof. done. Qed.

  Lemma memory_update_mrestrict_commutative
        (loc:locT) (locs: \powerset locT) (Hloc: loc !\in locs)
        (val:valueT) (mem:memT) :
    mrestrict (E:=valueT) locs (memory_update acc (Elt (ssingleton loc)) val mem)
    ≡ memory_update acc (Elt (ssingleton loc)) val (mrestrict (E:=valueT) locs mem).
  Proof.
    move: mem=> [mem|]; last done.
    apply: mem_eq_spec; [done|done|move=> loc'].
    rewrite memory_update_spec mem_mrestrict_spec.
    rewrite {1}/memory_update.
    rewrite {1}/mrestrict/memT_restricting_maplike/withtop_restricting_maplike.
    case: (ssingleton_spec loc' loc)=> [[Hin Heq2]|[Hin Heq2]].
    { usual_tac Hin; rewrite Heq2.
      rewrite Hloc.
      rewrite memory_update_spec Heq2.
      case: (_ && _); first reflexivity.
      rewrite mem_mrestrict_spec Hloc.
      reflexivity.
    }
    { move/Bool.negb_true_iff in Heq2.
      rewrite Heq2.
      rewrite mem_mrestrict_spec.
      case: (_ !\in _); last reflexivity.
      rewrite memory_update_spec Heq2.
      reflexivity.
    }
  Qed.

  Lemma memory_update_mrestrict_join_commutative
        (loc:locT) (locs: \powerset locT) (Hloc: loc !\in locs)
        (val:valueT) (mem mem':memT) :
    mrestrict_join (E:=valueT) locs (memory_update acc (Elt (ssingleton loc)) val mem) mem'
    ≡ memory_update acc (Elt (ssingleton loc)) val (mrestrict_join (E:=valueT) locs mem mem').
  Proof.
    move: mem=> [mem|]; last done.
    move: mem'=> [mem'|]; last done.
    apply: mem_eq_spec; [done|done|move=> loc'].
    rewrite memory_update_spec mem_mrestrict_join_spec.
    rewrite {1}/memory_update.
    rewrite {1}/mrestrict_join/mrestrict/memT_restricting_maplike/withtop_restricting_maplike.
    case: (ssingleton_spec loc' loc)=> [[Hin Heq2]|[Hin Heq2]].
    { usual_tac Hin; rewrite Heq2.
      rewrite Hloc.
      rewrite memory_update_spec Heq2.
      case: (_ && _); first reflexivity.
      rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin/withtop_join.
      rewrite mem_mrestrict_join_spec Hloc.
      reflexivity.
    }
    { move/Bool.negb_true_iff in Heq2.
      rewrite Heq2.
      rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin/withtop_join.
      rewrite mem_mrestrict_join_spec.
      case: (_ !\in _); last reflexivity.
      rewrite memory_update_spec Heq2.
      reflexivity.
    }
  Qed.

  Lemma memory_update_mrestrict_join_commutative_subset :
    forall (tlocs: Set! [locT]) (locs: \powerset locT)
           (Hloc: le (set_to_powerset tlocs) locs)
           (val:valueT) (mem mem':memT),
      mrestrict_join (E:=valueT) locs
                     (memory_update acc (Elt tlocs) val mem) mem'
      ≡
      memory_update acc (Elt tlocs) val
      (mrestrict_join (E:=valueT) locs mem mem').
  Proof.
    move=> tlocs locs Hloc val [mem|] [mem'|]; try done.
    apply mem_eq_spec; try done.
    move=> loc.
    rewrite memory_update_spec mem_mrestrict_join_spec.
    rewrite {1}/memory_update.
    rewrite {1}/mrestrict_join/mrestrict/memT_restricting_maplike
            /withtop_restricting_maplike.
    case Htlocsin: (loc !\in tlocs).
    { case Hlocsin: (loc !\in locs).
      { rewrite memory_update_spec Htlocsin.
        case (_ && _); first by exact:eq_refl.
        rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin
                /withtop_join.
        rewrite mem_mrestrict_join_spec Hlocsin.
        exact:eq_refl.
      }
      { elimtype False.
        rewrite/le/le_/morphism_poset/le/le_/bool_poset in Hloc.
        pose (Hloc loc) as Hlocin.
        simpl in Hlocsin.
        rewrite -> Hlocsin in Hlocin.
        rewrite/set_to_powerset in Hlocin.
        simpl in Hlocin.
        simpl in Htlocsin.
        rewrite Htlocsin in Hlocin.
        move/implyP in Hlocin.
        inversion Hlocin.
      }
    }
    { case Hlocsin: (loc !\in locs).
      { rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin
                /withtop_join.
        rewrite mem_mrestrict_join_spec Hlocsin.
        rewrite memory_update_spec Htlocsin.
        exact:eq_refl.
      }
      { rewrite [join (Elt _) (Elt _)]/join/join_/memT_hasJoin/withtop_hasJoin
                /withtop_join.
        rewrite mem_mrestrict_join_spec Hlocsin.
        exact:eq_refl.
      }
    }
  Qed.
  
  Lemma bind_mrestrict_commutative (argvals:list valueT) (params:list idT) (fid:fidT) (m:memT) (l: \powerset locT)
      (Hpin: forall i, List.In i params -> var_loc fid (local i) nil !\in l) :
      mrestrict (E:=valueT) l (sem_bind_params fid argvals params m) 
      ≡ sem_bind_params fid argvals params (mrestrict (E:=valueT) l m).
  Proof.
    elim: argvals params fid m l Hpin=> [|argval argvals IH] [|param params]; try reflexivity.
    move=> pidx mem locs Hparams.
    rewrite 2! sem_bind_params_simpl.
    rewrite IH.
    apply sem_bind_params_mor; try reflexivity.
    apply memory_update_mrestrict_commutative.
    { by apply: Hparams; left. }
    { by move=> i Hi; apply: Hparams; right. }
  Qed.

  Lemma bind_mrestrict_join_commutative (l: \powerset locT) (m m':memT) 
    (fid:fidT) (argvals:list valueT) params
    (Hpin: forall i, List.In i params -> var_loc fid (local i) nil !\in l) :
    mrestrict_join (E:=valueT) l (sem_bind_params fid argvals params m) m'
    ≡ sem_bind_params fid argvals params (mrestrict_join (E:=valueT) l m m').
  Proof.
    elim: argvals params m m' l Hpin => [|argval argvals IH] [|param params]; try reflexivity.
    move=> mem mem' locs Hparams.
    rewrite 2! sem_bind_params_simpl.
    rewrite IH.
    apply sem_bind_params_mor; try reflexivity.
    apply memory_update_mrestrict_join_commutative.
    { by apply: Hparams; left. }
    { by move=> i Hi; apply: Hparams; right. }
  Qed.
  
  Lemma mrestrict_join_bind_merge (l: \powerset locT) (m m':memT) 
    (fid:fidT) (argvals:list valueT) params
    (Hpin: forall i, List.In i params -> var_loc fid (local i) nil !\in l) :
    mrestrict_join (E:=valueT) l m (sem_bind_params fid argvals params m')
    ≡ mrestrict_join (E:=valueT) l m m'.
  Proof.
    elim: argvals params m m' Hpin=> [|argval argvals IH] [|param params] m m' Hpin; try exact: eq_refl.
    rewrite IH; last by move=> id Hid; apply Hpin; right.
    rewrite/memory_update.
    apply join_mor; first exact: eq_refl.
    move: m'=> [m'|]; last done.
    apply mem_eq_spec; [done|done|move=> loc].
    rewrite ? mem_mrestrict_spec.
    case Hin: (_ !\in _); last done.
    rewrite elements_singleton.
    rewrite ? [(Elt _)[_]!]/mlookup/memT_maplike/withtop_maplike.
    case: (is_strong _ _); rewrite map_madd_spec.
    { case Heq: (eqb _ _); last reflexivity.
      move/eqbP in Heq; usual_tac Heq.
      rewrite powerset_neg_spec in Hin.
      by rewrite Hpin in Hin; last by left.
    }
    { case Heq: (eqb _ _); last reflexivity.
      move/eqbP in Heq; usual_tac Heq.
      rewrite powerset_neg_spec in Hin.
      by rewrite Hpin in Hin; last by left.
    }
  Qed.
End SemAbsBlocks.
