(*******************************************************************)
(* colleting all strings from the program and the analysis results *)
(*******************************************************************)

let rec unions (ls : ('a PSet.t) list) : 'a PSet.t =
  match ls with
  | [] -> PSet.empty
  | hd::tl -> PSet.union hd (unions tl)

let unions_pset (s : ('a PSet.t) PSet.t) : 'a PSet.t =
  PSet.fold (fun a b -> PSet.union a b) s PSet.empty

let unions_array (a : ('a PSet.t) array) : 'a PSet.t =
  Array.fold_right (fun a b -> PSet.union a b) a PSet.empty

let collect_strs_opt (o: 'a option) (f: 'a -> 'b PSet.t) : 'b PSet.t =
  match o with
  | Some o1 -> f o1
  | None -> PSet.empty

let collect_strs_pair (x: 'a * 'b) 
    (collect_strs_a : 'a -> 'c PSet.t)
    (collect_strs_b : 'b -> 'c PSet.t) : 'c PSet.t =
  match x with
  | (a,b) -> PSet.union (collect_strs_a a) (collect_strs_b b)

let collect_strs_list (l:'a list) (f:'a -> 'b PSet.t) : 'b PSet.t =
  unions (List.map f l)

let collect_strs_pset (s: 'a PSet.t) (collect_strs_a : 'a -> 'b PSet.t) : 'b PSet.t =
  unions_pset (PSet.map collect_strs_a s)

let collect_strs_array (a:'a array) (f:'a -> 'b PSet.t) : 'b PSet.t =
  unions_array (Array.map f a)

let collect_strs_pmap (m:('a,'b) PMap.t)
    (collect_strs_a : 'a -> 'c PSet.t)
    (collect_strs_b : 'b -> 'c PSet.t) : 'c PSet.t =
  PMap.foldi
    (fun a b c -> unions [ collect_strs_a a; collect_strs_b b; c ])
    m PSet.empty
let collect_strs_pmap_key (m:('a,'b) PMap.t)
    (collect_strs_a : 'a -> 'c PSet.t) : 'c PSet.t =
  PMap.foldi (fun a b c -> PSet.union (collect_strs_a a) c ) m PSet.empty
let collect_strs_pmap_value (m:('a,'b) PMap.t)
    (collect_strs_b : 'b -> 'c PSet.t) : 'c PSet.t =
  PMap.fold (fun b c -> PSet.union (collect_strs_b b) c ) m PSet.empty

let collect_strs_string (s:string) : string PSet.t = PSet.singleton s
let collect_strs_string_list (sl:string list) : string PSet.t =
  collect_strs_list sl collect_strs_string

let rec collect_strs_exp (e:Syn.exp) : string PSet.t =
  match e with
  | Syn.Econst lit -> PSet.empty
  | Syn.Ebop (b,e1,e2) -> 
    PSet.union (collect_strs_exp e1) (collect_strs_exp e2)
  | Syn.Euop (u,e1) -> collect_strs_exp e1
  | Syn.EaddrOf lv -> collect_strs_lv lv
  | Syn.EvalueOf lv -> collect_strs_lv lv
  | Syn.Ecast (ct,e1,e2) -> 
    unions [ collect_strs_ctype ct
           ; collect_strs_exp e1
           ; collect_strs_exp e2 ]
  | Syn.EsizeOf e1 -> collect_strs_exp e1

and collect_strs_lv (lv:Syn.lv) : string PSet.t =
  match lv with
  | Syn.Lid x -> collect_strs_string x
  | Syn.Lderef e -> collect_strs_exp e
  | Syn.Larray (e1,e2) -> PSet.union (collect_strs_exp e1) (collect_strs_exp e2)
  | Syn.Lfield (e,x) -> PSet.union (collect_strs_exp e) (collect_strs_string x)

and collect_strs_ctype (ct:Syn.ctypeinfo) : string PSet.t =
  match ct with
  | (_,Syn.STRUCT s) -> collect_strs_string s
  | (_,Syn.ASTRUCT str_ct_list) -> collect_strs_str_ct_list str_ct_list
  | (_,Syn.UNION s) -> collect_strs_string s
  | (_,Syn.AUNION str_ct_list) -> collect_strs_str_ct_list str_ct_list
  | (_,Syn.POINTER ct1) -> collect_strs_ctype ct1
  | (_,Syn.ARRAY (ct1,e)) -> 
    PSet.union (collect_strs_ctype ct1) (collect_strs_exp e)
  | (_,Syn.FUNCTION ct1) -> collect_strs_ctype ct1
  | _ -> PSet.empty

and collect_strs_str_ct_list (str_ct_list:(string * Syn.ctypeinfo) list) : string PSet.t =
  collect_strs_list str_ct_list
    (fun x -> collect_strs_pair x collect_strs_string collect_strs_ctype)

let collect_strs_exp_list (e_list:Syn.exp list) : string PSet.t =
  collect_strs_list e_list collect_strs_exp

let collect_strs_position (p:Syn.position) : string PSet.t = 
  match p with
  | (_, _, s) -> collect_strs_string s

let collect_strs_tag (t:Syn.tag) : string PSet.t =
  match t with 
  | (scp,lbl,pos) ->
    unions [ collect_strs_string scp
           ; collect_strs_string lbl
           ; collect_strs_position pos ]

let rec collect_strs_fact (f:Syn.fact) : string PSet.t =
  match f with 
  | Syn.Pbin (_, e1, e2) -> 
    PSet.union (collect_strs_exp e1) (collect_strs_exp e2)
  | Syn.Pnot f1 -> collect_strs_fact f1

let collect_strs_lv_opt (lv_opt:Syn.lv option) : string PSet.t =
  collect_strs_opt lv_opt collect_strs_lv

let collect_strs_lv_list (lv_list:Syn.lv list) : string PSet.t =
  collect_strs_list lv_list collect_strs_lv

let collect_strs_alloc (a:Syn.alloc) : string PSet.t =
  match a with
  | Syn.Aarr (e1,e2) -> PSet.union (collect_strs_exp e1) (collect_strs_exp e2)
  | Syn.Arec f_list -> collect_strs_string_list f_list

let collect_strs_bcmd (cmd:Cfg.bcmd) : string PSet.t =
  match cmd with
  | Cfg.SET (lv,e) -> PSet.union (collect_strs_lv lv) (collect_strs_exp e)
  | Cfg.WEAKSET (lv,e) -> PSet.union (collect_strs_lv lv) (collect_strs_exp e)
  | Cfg.ESCAPE e -> collect_strs_exp e
  | Cfg.ASSUME p -> collect_strs_fact p
  | Cfg.ASSERT p -> collect_strs_fact p
  | Cfg.ALLOC (lv,a,sp) -> 
    PSet.union (collect_strs_lv lv) (collect_strs_alloc a)
  | Cfg.CHECK e -> collect_strs_exp e
  | Cfg.LIBCALL (lv_opt, e, e_list) ->
    unions [ collect_strs_lv_opt lv_opt
           ; collect_strs_exp    e
           ; collect_strs_exp_list e_list ]
  | Cfg.FREE e -> collect_strs_exp e
  | Cfg.REFINE (_, lv_list, _, _) -> collect_strs_lv_list lv_list
  | Cfg.REFINEO (f, _) -> collect_strs_fact f
  | Cfg.NOP -> PSet.empty

let collect_strs_bcmds (cmds:Cfg.bcmds) : string PSet.t =
  collect_strs_list cmds 
    (fun x -> collect_strs_pair x collect_strs_bcmd collect_strs_tag)

let collect_strs_block (b:Cfg.block) : string PSet.t =
  match b with
  | Cfg.CALL ((lv_opt, e, e_list), t, cmds) ->
    unions [ collect_strs_lv_opt   lv_opt
           ; collect_strs_exp      e
           ; collect_strs_exp_list e_list
           ; collect_strs_tag      t
           ; collect_strs_bcmds    cmds ]
  | Cfg.BLOCK cmds -> collect_strs_bcmds cmds

let collect_strs_blocks (bs:(Cfg.node, Cfg.block) PMap.t) : string PSet.t =
  collect_strs_pmap_value bs collect_strs_block

let collect_strs_locs (ls:Syn.locs) : string PSet.t =
  unions [ ls.Syn.pure_locals; ls.Syn.locals; ls.Syn.at; ls.Syn.ret ]

let collect_strs_cfg (c:Cfg.t) : string PSet.t =
  unions
    [ collect_strs_string      c.Cfg.pid 
    ; collect_strs_string_list c.Cfg.params
    ; collect_strs_blocks      c.Cfg.blocks
    ; collect_strs_locs        c.Cfg.locs
    ; collect_strs_tag         c.Cfg.tag ]

let collect_strs_graph_cfgs (gc:(Syn.pid, Cfg.t) PMap.t) : string PSet.t =
  collect_strs_pmap gc collect_strs_string collect_strs_cfg

let collect_strs_node (n:M5graph.node) : string PSet.t =
  match n with
  | (ctx,_) -> collect_strs_string ctx

let collect_strs_node_neighbors (nn:M5graph.node Vocabularies.neighbors) : string PSet.t =
  PSet.union
    (collect_strs_pset nn.Vocabularies.succ collect_strs_node)
    (collect_strs_pset nn.Vocabularies.pred collect_strs_node)

let collect_strs_graph_neighbors 
    (n:(M5graph.node, M5graph.node Vocabularies.neighbors) PMap.t) : string PSet.t =
  collect_strs_pmap n collect_strs_node collect_strs_node_neighbors

let collect_strs_node_set (ns:M5graph.node PSet.t) : string PSet.t =
  collect_strs_pset ns collect_strs_node

let collect_strs_graph_predcall
    (p:(M5graph.node, M5graph.node PSet.t) PMap.t) : string PSet.t =
  collect_strs_pmap p collect_strs_node collect_strs_node_set

let collect_strs_string_neighbors (sn:string Vocabularies.neighbors) : string PSet.t =
  PSet.union
    (collect_strs_pset sn.Vocabularies.succ collect_strs_string)
    (collect_strs_pset sn.Vocabularies.pred collect_strs_string)

let collect_strs_graph_calls
    (c:(string, string Vocabularies.neighbors) PMap.t) : string PSet.t =
  collect_strs_pmap c collect_strs_string collect_strs_string_neighbors

let collect_strs_graph_reaches 
    (r:(string, string Vocabularies.neighbors) PMap.t) : string PSet.t =
  collect_strs_graph_calls r

let collect_strs_graph_exits
    (e:(M5graph.node, M5graph.node PSet.t) PMap.t) : string PSet.t =
  collect_strs_graph_predcall e

let collect_strs_graph_indexof (i:(M5graph.node, int) PMap.t) : string PSet.t =
  collect_strs_pmap_key i collect_strs_node

let collect_strs_graph_nodeof (n:(int ,M5graph.node) PMap.t) : string PSet.t =
  collect_strs_pmap_value n collect_strs_node

let collect_strs_graph_domtree 
    (d:(M5graph.node ,M5graph.node) PMap.t) : string PSet.t =
  collect_strs_pmap d collect_strs_node collect_strs_node

let collect_strs_global_graph (g:M5graph.graph) : string PSet.t =
  unions
    [ collect_strs_graph_cfgs      g.M5graph.cfgs
    ; collect_strs_graph_neighbors g.M5graph.neighbors
    ; collect_strs_graph_predcall  g.M5graph.predcall
    ;                              g.M5graph.ctxs
    ; collect_strs_graph_calls     g.M5graph.calls
    ; collect_strs_graph_reaches   g.M5graph.reaches
    ;                              g.M5graph.loopheads
    ; collect_strs_graph_exits     g.M5graph.exits
    ; collect_strs_graph_indexof   g.M5graph.indexof
    ; collect_strs_graph_nodeof    g.M5graph.nodeof
    ; collect_strs_graph_domtree   g.M5graph.domtree ]

let collect_strs_addr1 (a:Dom.Addr1.t) : string PSet.t =
  match a with
  | Dom.Addr1.Variable id -> collect_strs_string id
  | Dom.Addr1.Dynamic id -> collect_strs_string id
  | Dom.Addr1.Null -> PSet.empty

let collect_strs_addrs (a:Dom.Addr1.t PSet.t) : string PSet.t =
  collect_strs_pset a collect_strs_addr1

let collect_strs_address (a:Dom.Address.t) : string PSet.t =
  match a with
  | Dom.Address.Top -> PSet.empty
  | Dom.Address.V addrs -> collect_strs_addrs addrs

let collect_strs_global_dump (g:Dom.Dump.t) : string PSet.t = 
  collect_strs_pmap g collect_strs_string collect_strs_address

let collect_strs_global_locs (locs:(Syn.pid,Syn.locs) PMap.t) : string PSet.t =
  collect_strs_pmap locs collect_strs_string collect_strs_locs

let collect_strs_global (g:Dom.global) : string PSet.t =
  unions
    [ collect_strs_global_graph g.Dom.graph
    ; collect_strs_global_dump  g.Dom.dump
    ; collect_strs_global_locs  g.Dom.locs ]

let collect_strs_context (c:Dom.Context.t) : string PSet.t =
  collect_strs_list c collect_strs_node

let collect_strs_proc (p:Dom.Proc.t) : string PSet.t =
  match p with
  | (s,s_list) -> 
    PSet.union (collect_strs_string s) (collect_strs_string_list s_list)

let collect_strs_closure (cls:Dom.Closure.t) : string PSet.t =
  match cls with
  | Dom.Closure.Top -> PSet.empty
  | Dom.Closure.V procs -> collect_strs_pset procs collect_strs_proc

let collect_strs_arrayblock (ab:Dom.ArrayBlock.t) : string PSet.t =
  match ab with
  | Dom.ArrayBlock.Top -> PSet.empty
  | Dom.ArrayBlock.V id_blk_map -> 
    collect_strs_pmap_key id_blk_map collect_strs_string

let collect_strs_value (v:Dom.Value.t) : string PSet.t =
  match v with
  | (_,cls,add,arrblk,_) ->
    unions 
      [ collect_strs_closure cls
      ; collect_strs_address add
      ; collect_strs_arrayblock arrblk ]


let collect_strs_mem (m:Dom.Mem.t) : string PSet.t =
  collect_strs_pmap m collect_strs_addr1 collect_strs_value

let collect_strs_state (s:Dom.State.t) : string PSet.t =
  collect_strs_pmap s collect_strs_context collect_strs_mem

let collect_strs_table (tbl:Dom.Table.t) : string PSet.t =
  collect_strs_pmap (tbl.Dom.Table.table) 
    collect_strs_node collect_strs_state

(*******************************************************)
(* generate str2pos and pos2str from collected strings *)
(*******************************************************)

let gen_mapping_btwn_str_pos (strs:string PSet.t) : 
    ((string, int) PMap.t * 
     (int, string) PMap.t) =
  let (str2pos,pos2str,_) = 
    PSet.fold 
      (fun s (str2pos, pos2str, n) -> 
        (PMap.add s n str2pos, PMap.add n s pos2str, n+1))
      strs (PMap.empty, PMap.empty, 1)
  in
  (str2pos, pos2str)

let collect_strs_du_label (lbl:GraphLib.DUGraph.Label.t) : string PSet.t =
  collect_strs_pset (lbl.GraphLib.DUGraph.Label.set) collect_strs_addr1

let collect_strs_dugraph_map 
    (dm: ((M5graph.node * M5graph.node), GraphLib.DUGraph.Label.t) PMap.t)
    : string PSet.t =
  collect_strs_pmap dm
    (fun x -> collect_strs_pair x collect_strs_node collect_strs_node)
    collect_strs_du_label

let collect_strs_du_nodeset (ns:GraphLib.DUGraph.NodeSet.t) : string PSet.t =
  collect_strs_pset (ns.GraphLib.DUGraph.NodeSet.set) collect_strs_node

let collect_strs_dugraph_succ
    (s:(M5graph.node,GraphLib.DUGraph.NodeSet.t) PMap.t) : string PSet.t =
  collect_strs_pmap s collect_strs_node collect_strs_du_nodeset

let collect_strs_dugraph_pred 
    (p:(M5graph.node,GraphLib.DUGraph.NodeSet.t) PMap.t) : string PSet.t =
  collect_strs_dugraph_succ p

let collect_strs_dugraph_inode (i:(M5graph.node, int) PMap.t) : string PSet.t =
  collect_strs_graph_indexof i

let collect_strs_dugraph_ivar (i:(Dom.Addr1.t, int) PMap.t) : string PSet.t =
  collect_strs_pmap_key i collect_strs_addr1

let collect_strs_dugraph_vari (v:Dom.Addr1.t array) : string PSet.t =
  collect_strs_array v collect_strs_addr1

let collect_strs_dugraph (dg:GraphLib.DUGraph.t) : string PSet.t =
  unions [ collect_strs_dugraph_map   dg.GraphLib.DUGraph.map
         ; collect_strs_dugraph_succ  dg.GraphLib.DUGraph.succ
         ; collect_strs_dugraph_pred  dg.GraphLib.DUGraph.pred
         ; collect_strs_dugraph_inode dg.GraphLib.DUGraph.inode
         ; collect_strs_dugraph_ivar  dg.GraphLib.DUGraph.ivar
         ; collect_strs_dugraph_vari  dg.GraphLib.DUGraph.vari ]

let collect_strs_workorder (wo:GraphLib.WorkOrder.t) : string PSet.t =
  collect_strs_pmap_key wo collect_strs_node

let collect_strs_headorder (h:(M5graph.node, int) PMap.t) : string PSet.t =
  collect_strs_dugraph_inode h

let collect_strs_loopheads (lh:M5graph.node PSet.t) : string PSet.t =
  collect_strs_node_set lh

let collect_strs_ssarua (du:Ssarua.t) : string PSet.t =
  unions [ collect_strs_dugraph du.Ssarua.dugraph
         ; collect_strs_workorder   du.Ssarua.order
         ; collect_strs_headorder du.Ssarua.headorder
         ; collect_strs_loopheads du.Ssarua.loopheads ]

let collect_strs_addr1_pset (a:Dom.Addr1.t PSet.t) : string  PSet.t =
  collect_strs_pset a collect_strs_addr1

let collect_strs_location (l:Pre.Location.t) : string PSet.t =
  unions [ collect_strs_addr1_pset l.Pre.Location.access
         ; collect_strs_addr1_pset l.Pre.Location.read
         ; collect_strs_addr1_pset l.Pre.Location.write ]

let collect_strs_access (a:Pre.Access.t) : string PSet.t =
  collect_strs_pmap a collect_strs_node collect_strs_location

let collect_strs_cgraph (cg:Pre.CallGraph.CGraph.t) : string PSet.t =
  Pre.CallGraph.CGraph.fold_vertex (fun v acc -> PSet.add v acc) cg PSet.empty

let collect_strs_dfs_num (dn:(string, int) PMap.t) : string PSet.t =
  collect_strs_pmap_key dn collect_strs_string

let collect_strs_callgraph (cg:Pre.CallGraph.t) : string PSet.t =
  unions [ collect_strs_cgraph      cg.Pre.CallGraph.graph
         ; collect_strs_string_list cg.Pre.CallGraph.dfs
         ; collect_strs_dfs_num     cg.Pre.CallGraph.dfs_num ]

let collect_strs_pid_type (pt:Pre.CallNode.pid_type) : string PSet.t =
  match pt with
  | Pre.CallNode.Lib p  -> collect_strs_string p
  | Pre.CallNode.User p -> collect_strs_string p

let collect_strs_pid_type_set (pts:Pre.CallNode.pid_type PSet.t) : string PSet.t =
  collect_strs_pset pts collect_strs_pid_type

let collect_strs_callnode (cn:Pre.CallNode.t) : string PSet.t =
  match cn with
  | (node2str_set, pid_type_set) ->
    PSet.union 
      (collect_strs_pmap node2str_set 
         collect_strs_node collect_strs_pid_type_set)
      (collect_strs_pid_type_set pid_type_set)

let collect_strs_pure_locals (pl:(string, Dom.Addr1.t PSet.t) PMap.t) : string PSet.t =
  collect_strs_pmap pl collect_strs_string collect_strs_addr1_pset

let collect_strs_return_addrs (ra:(string, Dom.Address.t) PMap.t) : string PSet.t =
  collect_strs_pmap ra collect_strs_string collect_strs_address

let collect_strs_return2sth (rs:(Dom.Addr1.t, M5graph.node PSet.t) PMap.t) 
    : string PSet.t =
  collect_strs_pmap rs collect_strs_addr1 collect_strs_node_set

let collect_strs_accessnodes (an:Pre.AccessEq.accessnodes) : string PSet.t =
  PSet.union 
    (collect_strs_node_set an.Pre.AccessEq.def)
    (collect_strs_node_set an.Pre.AccessEq.use)

let collect_strs_accesseq (ae:Pre.AccessEq.t) : string PSet.t =
  collect_strs_pmap ae collect_strs_addr1 collect_strs_accessnodes

let collect_strs_pre (p:Pre.preinfo) : string PSet.t =
  unions [ collect_strs_mem          p.Pre.mem
         ; collect_strs_access       p.Pre.access_original
         ; collect_strs_access       p.Pre.access_for_du
         ; collect_strs_access       p.Pre.access_for_globalssa
         ; collect_strs_addr1_pset   p.Pre.abslocs
         ; collect_strs_callgraph    p.Pre.callgraph
         ; collect_strs_callnode     p.Pre.callnode
         ; collect_strs_addr1_pset   p.Pre.variables
         ; collect_strs_addr1_pset   p.Pre.non_locals
         ; collect_strs_pure_locals  p.Pre.pure_locals 
         ; collect_strs_return_addrs p.Pre.return_addrs
         ; collect_strs_return2sth   p.Pre.return2uses
         ; collect_strs_return2sth   p.Pre.return2defs
         ; collect_strs_accesseq     p.Pre.simpledefs ]
