Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomZ.
Require Import DomMap.

Require Import Syntax.

Require Import DomPidx.
Require Import DomPidxSen.
Require Import DomConValue.

Set Implicit Arguments.


Section Nths.
  Lemma nth'_length' {A} (l l':list A) :
    nth' (length l) (l++l') = nth' 0 l'.
  Proof. by elim: l=> [|a l IH]. Qed.

  Lemma nth'_lt_length {A} n (l l':list A) (Hle: n < length l) :
    nth' n (l++l') = nth' n l.
  Proof.
    elim: l n Hle=> [|a l IH] [|n] Hle; rewrite/length/= in Hle *; try omega; try done.
    apply Lt.lt_S_n in Hle.
    exact: IH.
  Qed.
  
  Definition nth_fid {A B} (l:list (A*B)) (n:nat) (lastelt:A) : option A :=
    if eqb (length l) n
    then Some lastelt
    else if nth' n (rev l) is Some ab
         then Some ab.1
         else None.

  Lemma nth_fid_length {A B} (l:list (A*B)) (lastelt:A) :
    nth_fid l (length l) lastelt = Some lastelt.
  Proof. by rewrite/nth_fid eqb_refl. Qed.

  Global Add Parametric Morphism : (@nth_fid fidT bidT) with
      signature eq ==> eq ==> eq ==> eq
        as nth_mor.
  Proof.
    move=> x y Hxy z w Hzw a b Hab.
    usual_tac Hxy; usual_tac Hzw; usual_tac Hab.
    exact: eq_refl.
  Qed.

  Lemma nth_fid_subsume fid (bid:bidT) stack depth (default:fidT) :
    nth_fid ((fid,bid)::stack) depth default =
    if eqb (S (length stack)) depth
    then Some default
    else nth_fid stack depth fid.
  Proof.
    rewrite/nth_fid/=.
    case Heqb1: (eqb _ _); move/eqbP in Heqb1; first done.
    case Heqb2: (eqb (length stack) depth); move/eqbP in Heqb2.
    { usual_tac Heqb2.
      by rewrite <- rev_length; rewrite nth'_length'.
    }
    rewrite/eq/= in Heqb1 Heqb2.
    rewrite <- rev_length in Heqb1.
    rewrite <- rev_length in Heqb2.
    { elim: {stack} (rev stack) depth Heqb1 Heqb2=> [|top stack IH] [|depth]; rewrite/=; try done.
      move=> Heqb1 Heqb2.
      exploit IH.
      by contradict Heqb1; congr S; eauto.
      by contradict Heqb2; congr S; eauto.
      done.
    }
  Qed.

  Lemma nth_fid_some (stack:list (fidT*bidT)) depth (default:fidT) :
    (nth_fid stack depth default <> None) <->
    depth <= (length stack).
  Proof.
    rewrite/nth_fid.
    case Heqb1: (eqb (length stack)); move/eqbP in Heqb1.
    { usual_tac Heqb1.
      by split=> Hyp; try done; try omega.
    }
    { rewrite <- rev_length in Heqb1.
      rewrite <- rev_length.
      rewrite/eq/= in Heqb1.
      elim: {stack} (rev stack) depth Heqb1=> [|top stack IH] [|depth]; rewrite/=.
      { by try done; try omega. }
      { by move=> _; split=> Hyp; try done; try omega. }
      { by move=> _; split=> Hyp; try done; try omega. }
      { move=> Hyp; split=> Hyp'.
        apply: Le.le_n_S.
        apply IH; try done; try omega.
        apply IH; omega.
      }
    }
  Qed.

  Lemma nth_fid_none stack depth fid result
        (Hnth : nth_fid stack depth fid = Some result)
        (Hnone : List.find (fun (fb:fidT*bidT) => eqb result fb.1) stack = None) :
    depth = length stack /\ fid = result.
  Proof.
    rewrite/nth_fid in Hnth.
    case Heqb: (eqb _ _) in Hnth; first by move/eqbP in Heqb; inv Heqb; inv Hnth.
    apply find_none_rev in Hnone.
    elimtype False.
    elim: depth {stack Heqb} (rev stack) Hnth Hnone=> [|depth IH] [|top stack] Hnth Hnone; try done.
    { inv Hnth.
      by rewrite/= eqb_refl in Hnone.
    }
    { eapply IH; eauto.
      rewrite/= in Hnone.
      by case (eqb _ _) in Hnone.
    }
  Qed.
End Nths.


Section Memories.
Definition memT := locT ->> (option valueT).

Definition loc_wf (pidx:pidxT) (loc:locT) : Prop :=
  if loc is (inr (fid, _, depth), _)
  then eq (nth_fid pidx.2 depth (pifid pidx)) (Some fid)
  else True.

Global Add Parametric Morphism : loc_wf with
    signature eq ==> eq ==> eq as loc_wf_mor.
Proof.
  move=> x y Hxy; usual_tac Hxy.
  by move=> z w Hzw; usual_tac Hzw.
Qed.

Definition mem_wf (pidx:pidxT) (mem:memT) : Prop :=
  forall loc (Hloc: mem[loc]! <> (None:option valueT)),
    loc_wf pidx loc.

Global Add Parametric Morphism : mem_wf with
    signature eq ==> eq ==> le as mem_wf_le_mor.
Proof.
  move=> x y Hxy z w Hzw.
  usual_tac Hxy.
  refine (forall_implies _ _ _ _)=> loc.
  rewrite Hzw.
  by refine (forall_implies _ _ _ _)=> Hloc.
Qed.

Global Add Parametric Morphism : mem_wf with
    signature eq ==> eq ==> eq as mem_wf_mor.
Proof.
  move=> x y Hxy z w Hzw.
  apply/DomType.le_antisym.
  rewrite Hxy; rewrite Hzw.
  exact/DomType.eq_le/eq_refl.
  rewrite Hxy; rewrite Hzw.
  exact/DomType.eq_le/eq_refl.
Qed.
End Memories.
